#!/usr/bin/env python
#-*-coding: utf-8-*-

"""quartet.py - module for Quartet and quartet generation methods.

Function quartet_generator is the main function that takes care about
generating Quartet objects.
"""

from cogent.core.alignment import DenseAlignment
from numpy import array

__author__ = "Tomasz Puton"
__copyright__ = "Tomasz Puton"
__credits__ = ["Sandra Smit", "Kristian Rother", "Anthon Feenstra"]
__licence__ = "GPL"
__version__ = "0.1"
__maintainer__ = "Tomasz Puton"
__email__ = "tomasz.puton@gmail.com"
__status__ = "Production"

class QuartetGeneratorError(Exception):
    """Raised by quartet_generator when the Quartet cannot be created.
    """
    pass
    
class QuartetError(Exception):
    """Raised by Quartet upon providing incorrect input to __init__.
    """
    pass
    
class AverageQuartetError(Exception):
    """Raised by AverageQuartet upon providing incorrect input to update().
    """
    pass

class Quartet(object):
    """Quartet class represents 4 aligned sequences (= a quartet).
    
    Quartets are central part of the statistical geometry algorithm
    described first in the following papers:
    - PubMed ID: 2497522
    - PubMed ID: 3413065
    A nice review of the statistical geometry is also provided in the following
    paper:
    - PubMed ID: 9362556
    Therein a user can find more information about the algorithm itself.
    
    PUBLIC METHODS:
    - analyze(): calculates the characteristics of the Quartet. 
    Returns and puts them in the Params dict. Description and meanings of
    all characteristics that are derived from Quartet can be found in 
    the above-mentioned papers.
    
    - degap(): removes gaps from self._inp (only if it is DenseAlignment) and
    puts them in self._degapped_inp, which is used to re-initialize
    self.Quartet
    """
    # PROTRUSION_INDICES tell which sequence from a given alignment column ...
    # ... will be analyzed. The numbering starts from 0, so the position ...
    # ... number 1 has index 0. As there are always four sequences ...
    # ... in a column, 'alpha', 'beta', 'gamma' and 'delta' correspond to ...
    # ... all those sequences. These values are defined in e.g. 3413065 ... 
    # ... This dict is called PROTRUSION_INDICES because it allows ...
    # ... to calculate lenghts of protrusions from box to points ...
    # ... representing sequences in a sequence space (see the papers).
    PROTRUSION_INDICES = {'A|BCD' : 0, 'B|ACD' : 1, 'C|ABD' : 2, 'D|ABC' : 3}
    
    # INDICES is a set of all indices that we use for getting Quartet ...
    # ... positions. As there are 4 seqs, numbering starts from 0 ...
    # ... and ends at 3. INDICES is only used in _get_three_of_a_kind and ...
    # ... _get_pairwise_cong methods.
    INDICES = set([0,1,2,3])
    
    # BOX_INDICES defines the pairwise congruencies...
    # ... between positions in a ...
    # ... given column. They are best defined in 2497522, but in 3413065...
    # ... one can also find their explanation. E.g....:
    # ... 'AB' : sequences having indices 0 & 1 must be identical and ....
    # ... must differ from the two others (indices 2 & 3).
    # ... This dict is called BOX_INDICES because it allows to calculate...
    # ... lenghts of dimensions of the box (see the papers).
    BOX_INDICES = {'AB|CD' : (0, 1), 'AC|BD': (0, 2), 'AD|BC' : (0, 3)}
    
    REST_INDICES = [list(INDICES.difference([idx])) for idx in range(4)]
    REST_LIST = [[idx, REST_INDICES[idx][0], REST_INDICES[idx][1], \
            REST_INDICES[idx][2]] for idx in range(4)]
    
    TETRA_INDICES = {'AB|C|D' : (0, 1), 'AC|B|D' : (0, 2), 'AD|B|C' : (0, 3),
                     'BC|A|D' : (1, 2), 'BD|A|C' : (1, 3), 'CD|A|B' : (2, 3)}
    
    PAIRWISE_LIST = {}
    for idx_1 in range(4):
        for idx_2 in range(4):
            rest_of_indices = list(INDICES.difference([idx_1, idx_2]))
            PAIRWISE_LIST[(idx_1, idx_2)] = \
            [idx_1, idx_2, rest_of_indices[0], rest_of_indices[1]]
    
    def __init__(self, Aln):
        """Initializes new Quartet object.
        
        Aln: DenseAlignment object containing 4 ByteSequence objects
        (a quartet).
        
        Note: If the Quartet is not the DenseAlignment object 
        (containing different, than 4 number of ByteSequence objects),
        the QuartetError is raised.
        If you want to create Quartet objects by yourself (not with the
        help of quartet_generator), make sure DenseAlignment is correctly 
        initialized.
        """        
        if isinstance(Aln, DenseAlignment) and len(Aln.keys()) == 4:
            self.Quartet = array([byte_seq for byte_seq in Aln])
            self.Aln = Aln
        else:
            raise QuartetError("INCORRECT INPUT")
        
        # self.Params is a dict storing all values which are returned by ...
        # ... analyze method. These values are a Quartet ...
        # ... characteristics which are all described in the above-mentioned...
        # ... papers.
        self.Params = {}
        self._inp = Aln
        self.QuartetLen = len(self.Quartet)
        
    def __str__(self):
        """Returns str(self.Quartet).
        """
        return str(self.Quartet)

    @staticmethod
    def _get_identical_pair_indices(col):
        """Gets indices of the same items in a column.
            Used for calculating constituents of d3.
        
            col: a column (array) of characters
            
            Should return a tuple consisting of two indices.
        """
        if not hasattr(col, 'tolist'):
            raise QuartetError("Expecting array, got %s" % type(col))
            
        str_col = "".join(col.tolist())
        for i, item in enumerate(str_col):
            right_index = str_col.rfind(item)
            if i != right_index:
                return (i, right_index)
    
    def _get_d3_d4(self):
        """Returns d3 and d4 parameters at once.
        """
        tmp_counts_one_pair = {(0, 1) : 0, (0, 2) : 0, (0, 3) : 0, (1, 2) : 0,
                               (1, 3) : 0, (2, 3) : 0}
        one_pair, all_diff = 0, 0
        for col in self.Quartet:
            if len(set(col)) == 3:
                one_pair += 1       # We've found a column with one pair...
                tmp_counts_one_pair[self._get_identical_pair_indices(col)] += 1
                
            if len(set(col)) == 4:
                all_diff += 1       # ...& a column with all entries different.
                
        tetra_dict = Quartet.TETRA_INDICES
        # we iterate over list ['AB|C|D', 'AC|B|D', 'AD|B|C', 'BC|A|D',
        #                       'BD|A|C', 'CD|A|B']
        for idx in tetra_dict.keys():
            self.Params[idx] = tmp_counts_one_pair[tetra_dict[idx]] / \
                                float(self.QuartetLen)
            
        return [float(x) / self.QuartetLen for x in [one_pair, all_diff]]
        
    def _get_all_pos_equal(self):
        """Returns percentage of columns where all positions are equal.
        """
        all_equal = 0
        for col in self.Quartet:
            if len(set(col)) == 1:
                all_equal += 1
        return float(all_equal) / self.QuartetLen
        
    def _get_pairwise_cong(self, indices):
        """Returns percentage of columns with 2 pairs of equal entries.
        
        indices: a tuple of two integers
        
        Note: depending on the indices returns an appropriate number.
        IMPORTANT ! In order to get these values run the public method
        analyze. Don't try using this method by yourself 
        (unless you know what you're doing and you now meaning of 'indices')!
        """
        pairwise_cong = 0
        for col in self.Quartet:
            pairs_list = Quartet.PAIRWISE_LIST[indices]
            the_1st_chosen_one, the_2nd_chosen_one, rest_1, rest_2 = \
                col[pairs_list[0]] , col[pairs_list[1]], col[pairs_list[2]], col[pairs_list[3]]
            
            # Let's check for pairwise congruencies!
            if the_1st_chosen_one == the_2nd_chosen_one\
            and rest_1 == rest_2 and the_1st_chosen_one != rest_1\
            and the_2nd_chosen_one != rest_2:
                pairwise_cong += 1
        return float(pairwise_cong) / self.QuartetLen

    def _get_three_of_a_kind(self, idx):
        """Returns percentage of columns where three chracters are identical.
        
        Note: this value makes up the d1 parameter both in binary and 
        in quaternary space!!! Values obtained for four different indices:
        0, 1, 2, 3 correspond to 4 protrusions from a box (or some other shape
        in case of quaternary and n-ary space) to points representing 
        sequences.
        IMPORTANT ! In order to get these values run the public method analyze! 
        Don't try using this method by yourself 
        (unless you know what you're doing and you now the meaning of 'idx')!!!
        """
        nr_of_3_of_a_kind = 0
        for col in self.Quartet:
            rest_list = Quartet.REST_LIST[idx]
            the_chosen_one = col[rest_list[0]]
            rest = [col[rest_list[1]], col[rest_list[2]], col[rest_list[3]]]
            if the_chosen_one not in rest and len(set(rest)) == 1:
                nr_of_3_of_a_kind += 1
        return float(nr_of_3_of_a_kind) / self.QuartetLen
    
    def degap(self):
        """Removes gaps from self._inp, resets self.Quartet and self.QuartetLen
        
        Note: if Quartet is created by quartet_generator, then gaps are
        removed automatically and you don't have to use this method. However,
        if you create Quartet manually, you have to degap alignment it contains
        by using this method.
        """
        degapped_inp = self._inp.omitGapPositions(allowed_gap_frac=0.0)
        # Let's reset self.Quartet and self.QuartetLen
        self.Quartet = array([byte_seq for byte_seq in degapped_inp])
        self.QuartetLen = float(len(self.Quartet))
        
    def analyze(self):
        """Returns Quartet parameters and puts them in self.Params.
        """
        protr_dict = Quartet.PROTRUSION_INDICES
        tok = [self._get_three_of_a_kind(i) for i in range(4)]
        # we iterate over list ['A|BCD', 'B|ACD', 'C|ABD', 'D|ABC']:
        for idx in protr_dict.keys():
            self.Params[idx] = tok[protr_dict[idx]]
        
        box_dict = Quartet.BOX_INDICES
        # we iterate over list ['AB|CD', 'AC|BD', 'AD|BC']:
        for idx in box_dict.keys():
            self.Params[idx] = self._get_pairwise_cong(box_dict[idx])
            
        self.Params['d0'] = self._get_all_pos_equal()
        self.Params['d1'] = sum(self.Params[x] for x in protr_dict.keys())
        self.Params['d2'] = sum(self.Params[x] for x in box_dict.keys())

        sorted_dims = sorted([self.Params[idx] for idx in box_dict])
        self.Params['small_box'] = sorted_dims[0]
        self.Params['medium_box'] = sorted_dims[1]
        self.Params['large_box'] = sorted_dims[2]

        return self.Params

class BinaryQuartet(Quartet):
    """Inherits everything from Quaretet.
    """
    def __init__(self, Aln):
        """Initializes new BinaryQuartet object.
        
        Aln: DenseAlignment object containing 4 ByteSequence objects
        (a quartet).
        
        Note: If the Quartet is not the DenseAlignment object 
        (containing different, than 4 number of ByteSequence objects),
        the QuartetError is raised.
        If you want to create BinaryQuartet objects by yourself 
        (not with the help of quartet_generator), 
        make sure DenseAlignment is correctly initialized.
        """
        super(BinaryQuartet, self).__init__(Aln)
    
class QuaternaryQuartet(Quartet):
    """Inherits everything from Quaretet, except for analyze method.
    """
    def __init__(self, Aln):
        """Initializes new QuaternaryQuartet object.
        
        Aln: DenseAlignment object containing 4 ByteSequence objects
        (a quartet).
        
        Note: If the Quartet is not the DenseAlignment object 
        (containing different, than 4 number of ByteSequence objects),
        the QuartetError is raised.
        If you want to create QuaternaryQuartet objects by yourself 
        (not with the help of quartet_generator), 
        make sure DenseAlignment is correctly initialized.
        """
        super(QuaternaryQuartet, self).__init__(Aln)
    
    def analyze(self):
        """Returns QuaternaryQuartet parameters and puts them in self.Params.
        """
        super(QuaternaryQuartet, self).analyze()
        # d4 equals to 'A|B|C|D'
        self.Params['d3'], self.Params['d4'] = self._get_d3_d4()
        # By calling _get_d3_d4 parameters AB|C|D etc were calculated
        # therefore, we can now sort them and update self._small_tetra,
        # self._medium_tetra, self._large_tetra.
        
        tetra_dict = QuaternaryQuartet.TETRA_INDICES
        sorted_tetra = sorted([self.Params[idx] for idx in tetra_dict])
        self.Params['small_tetra'] = sorted_tetra[0]
        self.Params['medium_tetra'] = (sorted_tetra[2] + sorted_tetra[3]) / 2.0
        self.Params['large_tetra'] = sorted_tetra[-1]
        
        return self.Params

class AverageQuartet(object):
    """Keeps all the information extracted from Quartet.
    
    AverageQuartet keeps averaged characteristics of all Quartet objects, 
    that have been created so far.
    E.g. if 20000 Quartet objects have been created, AverageQuartet 
    contains averages of characteristics of 20000 Quartets
    (that is of 20000 alphas, 20000 betas, 20000 gammas and so on...).
    
    Values of AverageQuartet characteristics change when user calls update()
    method. This method takes existing Quartet object and updates 
    characteristics of the AverageQuartet. In order to calculate all the 
    averaged characteristics use calculateAverages() method.
    """
    def __init__(self):
        """Initializes new AverageQuartet object.
        """
        self.QuartetLen = 0.0
        
        self._AB = [0.0, 0]
        self._AC = [0.0, 0]
        self._AD = [0.0, 0]
        self._A_BCD = [0.0, 0]
        self._B_ACD = [0.0, 0]
        self._C_ABD = [0.0, 0]
        self._D_ABC = [0.0, 0]
        self._small_box = [0.0, 0]
        self._medium_box = [0.0, 0]
        self._large_box = [0.0, 0]
        self._d0 = [0.0, 0]
        self._d1 = [0.0, 0]
        self._d2 = [0.0, 0]
        
        self._AB_C_D = [0.0, 0]
        self._AC_B_D = [0.0, 0]
        self._AD_B_C = [0.0, 0]
        self._BC_A_D = [0.0, 0]
        self._BD_A_C = [0.0, 0]
        self._CD_A_B = [0.0, 0]
        self._small_tetra = [0.0, 0]
        self._medium_tetra = [0.0, 0]
        self._large_tetra = [0.0, 0]
        self._d3 = [0.0, 0]
        self._d4 = [0.0, 0]
        
        self._quartet_counter = 0
        
        # Interface
        self.SmallBox = 0
        self.MediumBox = 0
        self.LargeBox = 0
        self.SmallTetra = 0
        self.MediumTetra = 0
        self.LargeTetra = 0
        
        self.AB = None
        self.AC = None
        self.AD = None
        
        self.AB_C_D = None
        self.AC_B_D = None
        self.AD_B_C = None
        self.BC_A_D = None
        self.BD_A_C = None
        self.CD_A_B = None
        
        self.Protrusion = 0
        self.Alpha = None
        self.Beta = None
        self.Gamma = None
        self.Delta = None
        self.d0 = 0
        self.d1 = 0
        self.d2 = 0
        self.d3 = 0
        self.d4 = 0
        
    def reset(self):
        """Removes all of the stored characteristics.
        """
        self.__init__()
        
    def update(self, quartet):
        """Updates AverageQuartet with a Quartet params.
        
        quartet: Quartet object
        """
        if not isinstance(quartet, Quartet):
            raise AverageQuartetError("This is not a Quartet instance")
        elif len(quartet.Params) == 0:
            raise AverageQuartetError(\
            "Run \'analyze()\' of a Quartet first.")
        else:
            self._quartet_counter += 1
            ##
            self._AB = \
            [self._AB[0] + quartet.Params['AB|CD'], self._AB[1] + 1]

            self._AC = \
            [self._AC[0] + quartet.Params['AC|BD'], self._AC[1] + 1]
            
            self._AD = \
            [self._AD[0] + quartet.Params['AD|BC'], self._AD[1] + 1]
            
            ##
            self._d0 = \
            [self._d0[0] + quartet.Params['d0'], self._d0[1] + 1]
            
            self._d1 = \
            [self._d1[0] + quartet.Params['d1'], self._d1[1] + 1]
            
            self._d2 = \
            [self._d2[0] + quartet.Params['d2'], self._d2[1] + 1]
            
            ##
            self._small_box = \
            [self._small_box[0] + quartet.Params['small_box'], self._small_box[1] + 1]

            self._medium_box = \
            [self._medium_box[0] + quartet.Params['medium_box'], self._medium_box[1] + 1]
            
            self._large_box = \
            [self._large_box[0] + quartet.Params['large_box'], self._large_box[1] + 1]
                    
            if self.QuartetLen == 0.0:
                self.QuartetLen = quartet.QuartetLen
            else:
                if self.QuartetLen != quartet.QuartetLen:
                    raise AverageQuartetError("Quartets differ in lenght!")
                
            self._A_BCD = \
                [self._A_BCD[0] + quartet.Params['A|BCD'], self._A_BCD[1] + 1]
            
            self._B_ACD = \
                [self._B_ACD[0] + quartet.Params['B|ACD'], self._B_ACD[1] + 1]
            
            self._C_ABD = \
                [self._C_ABD[0] + quartet.Params['C|ABD'], self._C_ABD[1] + 1]
            
            self._D_ABC = \
                [self._D_ABC[0] + quartet.Params['D|ABC'], self._D_ABC[1] + 1]
            
    def calculateAverages(self):
        """Calculates averages of all Quartets' params.
        """
        self.Protrusion = (self._A_BCD[0] + self._B_ACD[0] + \
            self._C_ABD[0] + self._D_ABC[0]) / self._quartet_counter / 4
        
        self.d0 = self._d0[0] / self._d0[1]
        self.d1 = self._d1[0] / self._d1[1]
        self.d2 = self._d2[0] / self._d2[1]
        
        self.SmallBox = self._small_box[0] / self._small_box[1]
        self.MediumBox = self._medium_box[0] / self._medium_box[1]
        self.LargeBox = self._large_box[0] / self._large_box[1]
        
        if self._quartet_counter == 1:
            ##
            self.Alpha = self._A_BCD[0]
            self.Beta = self._B_ACD[0]
            self.Gamma = self._C_ABD[0]
            self.Delta = self._D_ABC[0]
            ##
            self.AB = self._AB[0]
            self.AC = self._AC[0]
            self.AD = self._AD[0]
                    
class AverageBinaryQuartet(AverageQuartet):
    """Keeps all the information extracted from BinaryQuartet.
    
    AverageBinaryQuartet keeps averaged characteristics of all 
    BinaryQuartet objects, that have been created so far.
    E.g. if 20000 BinaryQuartet objects have been created, AverageBinaryQuartet
    contains averages of characteristics of 20000 BinaryQuartet
    (that is of 20000 alphas, 20000 betas, 20000 gammas and so on...).
    
    Values of AverageBinaryQuartet characteristics change when user calls
    updateBinary() method. This method takes existing BinaryQuartet object and 
    updates characteristics of the AverageBinaryQuartet. In order to calculate
    all the averaged characteristics use calculateBinaryAverages() method.
    """
    def __init__(self):
        """Initializes new AverageBinaryQuartet object.
        """
        super(AverageBinaryQuartet, self).__init__()
        
    def update(self, quartet):
        """Updates AverageBinaryQuartet with a BinaryQuartet params.
        
        quartet: BinaryQuartet object
        """
        super(AverageBinaryQuartet, self).update(quartet)
        
    def calculateAverages(self):
        """Calculates averages of all BinaryQuartets' params.
        """
        super(AverageBinaryQuartet, self).calculateAverages()

class AverageQuaternaryQuartet(AverageQuartet):
    """Keeps all the information extracted from QuaternaryQuartet.
    
    AverageQuaternaryQuartet keeps averaged characteristics of all 
    QuaternaryQuartet objects, that have been created so far.
    E.g. if 20000 QuaternaryQuartet objects have been created, 
    AverageQuaternaryQuartet contains averages of characteristics of 
    20000 AverageQuaternaryQuartet (that is of 20000 alphas, 20000 betas, 
    20000 gammas and so on...).
    
    Values of AverageQuaternaryQuartet characteristics change when user calls
    updateQuaternary() method. This method takes existing QuaternaryQuartet 
    object and updates characteristics of the AverageQuaternaryQuartet. 
    In order to calculate all the averaged characteristics use 
    calculateQuaternaryAverages() method.
    """
    def __init__(self):
        """Initializes new AverageQuaternaryQuartet object.
        """
        super(AverageQuaternaryQuartet, self).__init__()
    
    def update(self, quartet):
        """Updates AverageQuaternaryQuartet with a QuaternaryQuartet params.
        
        quartet: BinaryQuartet object
        """
        super(AverageQuaternaryQuartet, self).update(quartet)
        self._AB_C_D = [self._AB_C_D[0] + quartet.Params['AB|C|D'],
                        self._AB_C_D[1] + 1]
        self._AC_B_D = [self._AC_B_D[0] + quartet.Params['AC|B|D'],
                        self._AC_B_D[1] + 1]
        self._AD_B_C = [self._AD_B_C[0] + quartet.Params['AD|B|C'],
                        self._AD_B_C[1] + 1]
        self._BC_A_D = [self._BC_A_D[0] + quartet.Params['BC|A|D'],
                        self._BC_A_D[1] + 1]
        self._BD_A_C = [self._BD_A_C[0] + quartet.Params['BD|A|C'],
                        self._BD_A_C[1] + 1]
        self._CD_A_B = [self._CD_A_B[0] + quartet.Params['CD|A|B'],
                        self._CD_A_B[1] + 1]
        self._d3 = [self._d3[0] + quartet.Params['d3'], self._d3[1] + 1]
        self._d4 = [self._d4[0] + quartet.Params['d4'], self._d4[1] + 1]
        
        self._small_tetra = [self._small_tetra[0] + \
                           quartet.Params['small_tetra'], \
                           self._small_tetra[1] + 1]

        self._medium_tetra = [self._medium_tetra[0] + \
                              quartet.Params['medium_tetra'], \
                              self._medium_tetra[1] + 1]
        
        self._large_tetra = [self._large_tetra[0] + \
                             quartet.Params['large_tetra'], \
                             self._large_tetra[1] + 1]
        
    def calculateAverages(self):
        """Calculates averages of all QuaternaryQuartets' params.
        """
        super(AverageQuaternaryQuartet, self).calculateAverages()
        
        self.SmallTetra = self._small_tetra[0] / self._small_tetra[1]
        self.MediumTetra = self._medium_tetra[0] / self._medium_tetra[1]
        self.LargeTetra = self._large_tetra[0] / self._large_tetra[1]
        
        self.d0 = self._d0[0] / self._d0[1]
        self.d3 = self._d3[0] / self._d3[1]
        self.d4 = self._d4[0] / self._d4[1]
        
        if self._quartet_counter == 1:
            self.AB_C_D = self._AB_C_D[0]
            self.AC_B_D = self._AC_B_D[0]
            self.AD_B_C = self._AD_B_C[0]
            self.BC_A_D = self._BC_A_D[0]
            self.BD_A_C = self._BD_A_C[0]
            self.CD_A_B = self._CD_A_B[0]

def select_from_one_DA(items, n = 4):
    """Returns list of unique quartets from 1 DenseAlignment object.
    
    items: list of ByteSequence objects from DenseAlignment.
    
    n: number of unique elements to select from a set of elements.
    
    Note: If n=4, selects unique quartets from a list of ByteSequence objects.
    It's a Python generator, so use method .next() to retrieve quartet!
    """
    if n == 0: yield []
    else:
        for i in xrange(len(items)):
            for cc in select_from_one_DA(items[i + 1:], n - 1):
                yield [items[i]] + cc
            
def select_from_four_DAs(items):
    """Returns list of unique quartets from 4 DenseAlignment objects.
    
    items: list of 4 DenseAlignment objects containing ByteSequence objects.
    
    Note: What it returns is a list of lists 
    (each element on a sublist is treated as a string).
    It's a Python generator, so use method .next() to retrieve next list!
    """
    # Every item in list 'items' in DenseAlignment object!
    for i in items[0].Seqs:     # Seqs is a list of ByteSequence objects...
        for j in items[1].Seqs:     # ...from a DenseAlignment!
            for k in items[2].Seqs:
                for l in items[3].Seqs:
                    # We convert ByteSequence (represented by 'x' below)...
                    # ...to list, because names of seqs in ByteSequence...
                    # ...are not unique, and therefore when one tries to...
                    # ...create a DenseAlignment object of them...
                    # ...error is raised!
                    yield [list(x) for x in [i, j, k, l]]

def check_input(input_):
    """Returns a list of DenseAlignments or raises QuartetGeneratorError.
    
    input_: a list created from input of quartet_generator function.
    """
    if type(input_) != type(list(DenseAlignment(['1','1']))):
        raise QuartetGeneratorError("INVALID INPUT")
    
    list_of_DenseAlignments = []
    for aln in input_:
        if isinstance(aln, DenseAlignment):
            list_of_DenseAlignments.append(aln)
    
    if len(list_of_DenseAlignments) not in (1, 4):
        raise QuartetGeneratorError("INVALID INPUT")
    else:
        return list_of_DenseAlignments
    
def quartet_generator\
(aln_A = None, aln_B = None, aln_C = None, aln_D = None, mode='B'):
    """Return Quartets generated from aln_A OR aln_A & aln_B & aln_C & aln_D.
    
    aln_A: DenseAlignmnet object, by default None
    aln_B: DenseAlignmnet object, by default None
    aln_C: DenseAlignmnet object, by default None
    aln_D: DenseAlignmnet object, by default None
    mode: 'B' for binary sequences, 'Q' for quaternary sequences
    frac: fraction of quartets to analyze, by default 1.0 (all)
    
    Note: It's a Python generator. Use method .next() to retrieve next Quartet.
    Input is 1 or 4 DenseAlignment objects which contain ByteSequence objects. 
    If 1 DenseAlignment object is passed, quartet_generator makes all quartets
    from its ByteSequence objects.
    If 4 DenseAlignment obejcts are passed, quartet_generator makes all 
    quartets taking one ByteSequence from each set.
    
    Setting frac lower than 1.0 will speed the calculations. E.g. frac equal to 
    0.01 will be 100 times faster than regular analysis.
    """
    input_ = [aln_A, aln_B, aln_C, aln_D]
    list_of_DenseAlignments = check_input(input_)
    
    if mode == 'B' or mode == 'b':
        quartet = BinaryQuartet
    if mode == 'Q' or mode == 'q':
        quartet = QuaternaryQuartet
    
    if len(list_of_DenseAlignments) == 1:
        # Let's make quartets from 1 DenseAlignment
        da = list_of_DenseAlignments[0].omitGapPositions(allowed_gap_frac=0.0)
        if len(da.Seqs) < 4:
            raise QuartetGeneratorError("Less, than 4 seqs in input!")
        if len(da) == 0:
            raise QuartetGeneratorError(\
            "No positions left in the alignment after gap removal!")
        
        for four_seqs in select_from_one_DA(\
        [da.getSeq(b) for b in da.keys()], 4):
            yield quartet(DenseAlignment(four_seqs))

    elif len(list_of_DenseAlignments) == 4:
        # Let's make quartets from 4 DenseAlignments
        tmp = list_of_DenseAlignments[0]
        for dense_aln in list_of_DenseAlignments[1:]:
            try:
                tmp = tmp.addSeqs(dense_aln)
            except ValueError, error:
                raise QuartetGeneratorError(error)
        
        names_1 = list_of_DenseAlignments[0].getSeqNames()
        names_2 = list_of_DenseAlignments[1].getSeqNames()
        names_3 = list_of_DenseAlignments[2].getSeqNames()
        names_4 = list_of_DenseAlignments[3].getSeqNames()
        
        ungapped = tmp.omitGapPositions(allowed_gap_frac=0.0)
        if len(ungapped) == 0:
            raise QuartetGeneratorError(\
        "No positions left in the alignment after gap removal!")
        
        bs_1 = [] # list of ByteSequence objects from the 1st DenseAlignment...
                  # ... from list_of_DenseAlignments
        bs_2 = []
        bs_3 = []
        bs_4 = []      
        
        for name in names_1:
            bs_1.append(ungapped.getSeq(name))
            
        for name in names_2:
            bs_2.append(ungapped.getSeq(name))
            
        for name in names_3:
            bs_3.append(ungapped.getSeq(name))
            
        for name in names_4:
            bs_4.append(ungapped.getSeq(name))
        
        for four_seqs in select_from_four_DAs(\
        [DenseAlignment(bs_1), DenseAlignment(bs_2), \
        DenseAlignment(bs_3), DenseAlignment(bs_4)]):
            yield quartet(DenseAlignment(four_seqs))
