#!/usr/bin/env python
"""
sglib.py - statistical geometry algorithm.

This method was introduced by Manfred Eigen in PNAS in 1988.
For the description of the method please refer to two articles:
- PubMed ID: 2497522
- PubMed ID: 3413065
"""

from cogent.core.alignment import DenseAlignment
from cogent.core.moltype import BYTES
from cogent.core.sequence import ByteSequence
from cogent.util.transform import make_trans

__author__ = "Tomasz Puton"
__copyright__ = "Tomasz Puton"
__credits__ = ["Sandra Smit", "Kristian Rother", "Anthon Feenstra"]
__licence__ = "GPL"
__version__ = "1.0"
__maintainer__ = "Tomasz Puton"
__email__ = "tomasz.puton@gmail.com"
__status__ = "Production"

RNA_BINARY = make_trans('UCAG','1100',default='-')
RNA_QUATERNARY = make_trans('UCAG','1234', default='-')
DNA_BINARY = make_trans('TCAG','1100',default='-')
DNA_QUATERNARY = make_trans('TCAG','1234', default='-')

class TranslationError(Exception):
    pass

def translate_lines(lines, codebook):
    """Returns DenseAlignment with translation.

    lines: ideally the cogent.parse.MinimalFastaParser object initialized
    with some FASTA file.
    But basically everything that can be iterated (list, dict, open files,
    etc), provided it has the same 'structure' as the output of 
    cogent.parse.MinimalFastaParser.next() 
    (eg. list of tuples ('>id', 'ACGU...')).
    
    codebook: a translation table made by cogent.util.transform.make_trans
    
    Note: Translates a record with sequence into BYTES record.
    Translates lines with aligned sequences.
    """
    
    translated_recs_list = []
    for record in lines:
        translation = record[1].translate(codebook)
        translated_recs_list.append(ByteSequence(translation, Name=record[0]))
    
    if len(set([len(bs) for bs in translated_recs_list])) != 1:
        raise TranslationError("Seqs are unequal. Are they aligned?!")
    
    return DenseAlignment(data = translated_recs_list, MolType = BYTES)

def translate_aln(aln, codebook):
    """Returns DenseAlignment with translation.
    
    aln: DenseAlignment object to be translated. Should contain cogent
    Sequence objects e.g. RnaSequence, ProteinSequence, DnaSequence objects.
    
    codebook: a translation table made by cogent.util.transform.make_trans
    
    Note: Translates a record with sequence into BYTES record.
    """
        
    translated_recs_list = []
    if not isinstance(aln, DenseAlignment):
        raise TranslationError("Valid input is DenseAlignment object.")
    else:
        for seq in aln.iterSeqs():
            translation = str(seq).translate(codebook)
            translated_recs_list.append(\
            ByteSequence(translation, Name=seq.getName()))
            
        return DenseAlignment(translated_recs_list)


#class SG(object):
#    def __init__(self, matrix):
#        """
#        Colects data and calculates the type of divergance.
#        
#        Argument is:
#        - binary_matrix
#        or
#        - quaternary_matrix
#        """
#        self.storage = Storage()
#        self.matrix = matrix
#        
#    def __add_quartet__(self, data):
#        """
#        Adds quartet. It should not be normally used.
#        Initialize SG object with matrix and process it instead.
#        """
#        self.last_quartet_added = Quartet(data)
#        self.last_quartet_added.store(self.storage)
#    
#    def __mix_columns__(self):
#        """
#        It should be run on large datasets only. It should be used only
#        when randomization has proceeded so much, that it is impossible
#        to distnguish between trees and bundles.
#        
#        It mixes columns only, rows remain unchanged.
#        
#        If the...
#        ...parameter average(large) + average(small) - 2 * average(medium)
#        decreases toward zero with increasing...
#        ...mixing, and if this decrease is significant as compared to its...
#        ...fluctuations, residual tree-likeness was present. (Before it...
#        ...was destroyed by mixing.)
#        """
#        pass
#        
#    def process_matrix(self):
#        """
#        Processes matrix (regardless whether it is binary or quaternary).
#        What does it do?
#        1. Reads matrix
#        2. Creates all quartets of sequences
#        """
#        pass
#         
#    def get_average_small(self):
#        """
#        Gets average value of parameter 'small'.
#        """
#        return self.storage.average_small
#
#    def get_average_medium(self):
#        """
#        Gets average value of parameter 'medium'.
#        """
#        return self.storage.average_medium
#
#    def get_average_large(self):
#        """
#        Gets average value of parameter 'large'.
#        """
#        return self.storage.average_large
#
#    def get_average_alpha(self):
#        """
#        Gets average value of parameter 'alpha'.
#        """
#        return self.storage.average_alpha
#        
#    def get_average_beta(self):
#        """
#        Gets average value of parameter 'beta'.
#        """
#        return self.storage.average_beta
#
#    def get_average_gamma(self):
#        """
#        Gets average value of parameter 'gamma'.
#        """
#        return self.storage.average_gamma
#        
#    def get_average_delta(self):
#        """
#        Gets average value of parameter 'delta'.
#        """
#        return self.storage.average_delta
#        
#class BinarySG(object):
#    def __init__(self, path, molecule):
#        """
#        Runs binary statistical geometry on RNA or protein sequences.
#        
#        Class with high level of abstraction - designed in order to make 
#        user's work smoother.
#        """
#        self.path = path
#        self.molecule = molecule
#        
#    def run(self):
#        """
#        Ths function should run the algorithm and return Results object.
#        """
#        results = None
#        alignment = Alignment(self.path)
#        
#        if self.molecule == 'RNA':
#            self.results = RNA_SG(alignment, 'BINARY')
#        if self.molecule == 'PROTEIN':
#            self.results = PROTEIN_SG(alignment, 'BINARY')
#    
#    def get_box_dimensions(self):
#        """
#        This function should return box dimensions.
#        """
#        return results.box_dimensions
#        
#    def get_box_protrusions(self):
#        """
#        This function should return average protrusions from box to
#        groups A, B, C and D e.g. Archea, Bacteria, Eukarya and mitochondria.
#        """
#        return results.box_protrusions
#        
#    def get_type_of_divergence(self):
#        """
#        This function should return type of divergance of analysed sequences.
#        """
#        return results.type_of_divergance
#            
#class QuaternarySG(object):
#    def __init__(self):
#        """
#        Runs binary statistical geometry on RNA or protein sequences.
#        
#        Class with high level of abstraction - designed in order to make 
#        user's work smoother.
#             
#        Let's leave this class empty fot now and focus on BinarySG.
#        It might be one of the possible future improvements to the program.
#        Similarly to RNA_SG it prepares an input to an object of 'SG' class.
#        """
#        pass
        
