#!/usr/bin/env python
#-*-coding: utf-8-*-

from __future__ import division

from numpy import array
from cogent.util.unit_test import TestCase, main
from cogent.core.alignment import DenseAlignment
from cogent.core.sequence import ByteSequence
from cogent.core.sequence import RnaSequence
from cogent.core.alignment import Alignment
from sglib.quartet import Quartet, BinaryQuartet, QuaternaryQuartet,\
QuartetGeneratorError, quartet_generator, check_input, select_from_one_DA, \
select_from_four_DAs, QuartetError, AverageQuartet, AverageBinaryQuartet,\
AverageQuaternaryQuartet, AverageQuartetError

class TestCaseBugFix(TestCase):
    """Custom assertEqualItems, which is used here for testing, because
        assertEqualItems from TestCase class from module cogent.utils.unit_tests
        does not handle comparing iterables with floats correctly!
        Such comparisons are critical in testing.
        
        The problem occured in PyCogent 1.5.0, no idea whether it occurs in
        next releases.
        
        Rounding to 7 digits after comma is used.
    """
    
    def assertEqualItems(self, observed, expected, msg=None, places=7):
        """Fail if the two items contain unequal elements"""
        obs_items = list(observed)
        exp_items = list(expected)
        if len(obs_items) != len(exp_items):
            raise self.failureException, \
            (msg or 'Observed and expected are different lengths: %s and %s' \
            % (len(obs_items), len(exp_items)))
            
        obs_items.sort()
        exp_items.sort()
        for index, (obs, exp) in enumerate(zip(obs_items, exp_items)):
            if obs != exp:
                if type(obs) == float and type(exp) == float:           # FIX
                    rounded_obs = round(obs, places)                    # FIX
                    rounded_exp = round(exp, places)                    # FIX
                    if rounded_obs != rounded_exp:                      # FIX
                        error_msg = \
                            'Compared floats (observed: %f vs expected: %f) ' %\
                            (obs, exp)                                  # FIX
                        error_msg += 'but they differed in more than %d places'\
                            % places                                    # FIX
                        raise self.failureException, (msg or error_msg) # FIX
                else:
                    raise self.failureException, \
                    (msg or 'Observed %s and expected %s at sorted index %s' \
                    % (obs, exp, index))

class QuartetTests(TestCaseBugFix):
    """Tests for Quartet object.
    """
    def setUp(self):
        """Creates environment for all tests of BinaryQuartet object.
        """
        b1 = ByteSequence('110010101', Name='b1')
        b2 = ByteSequence('110010000', Name='b2')
        b3 = ByteSequence('010100001', Name='b3')
        b4 = ByteSequence('110011111', Name='b4')
        b5 = ByteSequence('111111110', Name='b5')
        b6 = ByteSequence('011000010', Name='b6')
        
        self.four_byte_seqs_DIGITS = [b1, b2, b3, b4]
        self.REF_four_byte_seqs_DIGITS = \
array([['1', '1', '0', '1'],
       ['1', '1', '1', '1'],
       ['0', '0', '0', '0'],
       ['0', '0', '1', '0'],
       ['1', '1', '0', '1'],
       ['0', '0', '0', '1'],
       ['1', '0', '0', '1'],
       ['0', '0', '0', '1'],
       ['1', '0', '1', '1']], 
      dtype='|S1')
        
        self.valid_quartet_DIGITS = DenseAlignment(self.four_byte_seqs_DIGITS)
        self.invalid_quartet_DIGITS = DenseAlignment([b1, b2, b3, b4, b5])
        
        self.q1 = Quartet(self.valid_quartet_DIGITS)
        self.q2 = Quartet(DenseAlignment([b2,b3,b4,b5]))
        self.q3 = Quartet(DenseAlignment([b3,b4,b5,b6]))

        self.q_gaps = Quartet(DenseAlignment([b5, b6, b4,\
        ByteSequence('0110-----', Name='gaps')]))
        
    ## THIS SINGLE TEST BELOW CHECKS WHAT HAPPENS IF INPUT IS INVALID
    def test_init(self):
        """BinaryQuartet.__init__: should set self.Quartet & raise QuartetError
        """
        #
        BS = ByteSequence
        RS = RnaSequence
        DA = DenseAlignment
        #
        
        # We test here what happens if input is correct ...
        
        self.assertEqualItems(self.q1.Quartet.tolist(),\
        self.REF_four_byte_seqs_DIGITS.tolist())
        
        # ... and whether QuartetError is raised when it is incorrect.
        self.assertRaises(QuartetError, Quartet, self.invalid_quartet_DIGITS)
        
        self.assertRaises(QuartetError, Quartet, BS('111'))
        
        self.assertRaises(QuartetError, Quartet, \
                          [BS('1'), BS('1'), BS('1'), BS('1')])
        
        self.assertRaises(QuartetError, Quartet, 'AUGC')
        
        self.assertRaises(QuartetError, Quartet, [Alignment('U'),
                                                  Alignment('C'),
                                                  Alignment('G'),
                                                  Alignment('A')])
        
        self.assertRaises(QuartetError, Quartet, \
                          DA([BS('1'), BS('1'), BS('1')]))
        
        self.assertRaises(QuartetError, Quartet, DA([BS('1'), BS('1')]))
        
        self.assertRaises(QuartetError, Quartet, self.invalid_quartet_DIGITS)
        
        self.assertRaises(QuartetError, Quartet, BS('111'))
        
        self.assertRaises(QuartetError, Quartet, [BS('1'), BS('1'), BS('1'),
                                                  BS('1')])
        
        self.assertRaises(QuartetError, Quartet, 'ACUGT')
        
        self.assertRaises(QuartetError, Quartet, Alignment('ACUGT'))
        self.assertRaises(QuartetError, Quartet, Alignment('-----'))
        
        self.assertRaises(QuartetError, Quartet,[Alignment('U'),
                                                 Alignment('C'),
                                                 Alignment('G'),
                                                 Alignment('A')])
        
        self.assertRaises(QuartetError, Quartet, [Alignment('B'),
                                                  Alignment('B'),
                                                  Alignment('-'),
                                                  Alignment('A')])
        
    def test_str(self):
        """Quartet.__str__: should return string representation of Quartet
        """
        exp = str(self.REF_four_byte_seqs_DIGITS)       # DenseAlignment object
        obs = str(self.q1)                              # Quartet object 
        
        self.assertEqual(obs, exp)  # Although the objects are different, the
                                    # output should be the same, because
                                    # self.q1 was initialized with
                                    # self.valid_quartet_DIGITS.
        
    def test_get_identical_pair(self):
        """Quartet._get_identical_pair_indices should work correctly
        """
        case_AB_C_D = (array(['2', '2', '4', '1']), (0, 1))
        case_AC_B_D = (array(['A', '4', 'A', '1']), (0, 2))
        case_AD_B_C = (array(['1', '2', '3', '1']), (0, 3))
        case_BC_A_D = (array(['2', '1', '1', '4']), (1, 2))
        case_BD_A_C = (array(['4', 'x', '1', 'x']), (1, 3))
        case_CD_A_B = (array(['a', 'y', 'x', 'x']), (2, 3))
        
        self.assertEqual(self.q1._get_identical_pair_indices(case_AB_C_D[0]),
                         case_AB_C_D[1])
        self.assertEqual(self.q1._get_identical_pair_indices(case_AC_B_D[0]),
                         case_AC_B_D[1])
        self.assertEqual(self.q1._get_identical_pair_indices(case_AD_B_C[0]),
                         case_AD_B_C[1])
        self.assertEqual(self.q1._get_identical_pair_indices(case_BC_A_D[0]),
                         case_BC_A_D[1])
        self.assertEqual(self.q1._get_identical_pair_indices(case_BD_A_C[0]),
                         case_BD_A_C[1])
        self.assertEqual(self.q1._get_identical_pair_indices(case_CD_A_B[0]),
                         case_CD_A_B[1])
        
        self.assertRaises(QuartetError, self.q1._get_identical_pair_indices, [])
        self.assertRaises(QuartetError, self.q1._get_identical_pair_indices, '')
        self.assertRaises(QuartetError, self.q1._get_identical_pair_indices, \
                          None)
        
        case_invalid = (array(['2', '3', '4', '1']), None)
        self.assertEqual(self.q1._get_identical_pair_indices(case_invalid[0]),
                         case_invalid[1])
        
    def test_get_d3_d4_binary_space(self):
        """Quartet._get_d3_d4 should return 0, 0 in case of binary space
        """
        d3, d4 = self.q1._get_d3_d4()
        self.assertFloatEqual(d3, 0)
        self.assertFloatEqual(d4, 0)
        self.assertFloatEqual(self.q1.Params['AB|C|D'], 0)
        self.assertFloatEqual(self.q1.Params['AC|B|D'], 0)
        self.assertFloatEqual(self.q1.Params['AD|B|C'], 0)
        self.assertFloatEqual(self.q1.Params['BC|A|D'], 0)
        self.assertFloatEqual(self.q1.Params['BD|A|C'], 0)
        self.assertFloatEqual(self.q1.Params['CD|A|B'], 0)
        
    def test_get_all_pos_equal(self):
        """Quartet._get_all_pos_equal: should return % of cols of == pos
        """
        qb = self.q1
        qb.analyze()
        self.assertAlmostEqual(qb._get_all_pos_equal(), 0.222222, places=6)
        
    def test_get_three_of_a_kind(self):
        """Quartet._get_three_of_a_kind: returns cols' % with 3 eq pos.
        """
        self.assertEqual(self.q1._get_three_of_a_kind(0), 0)   # - A|BCD
        
        self.assertFloatEqual(\
        self.q1._get_three_of_a_kind(1), 0.1111111111111111)   # - B|ACD
        
        self.assertFloatEqual(\
        self.q1._get_three_of_a_kind(2), 0.3333333333333333)   # - C|ABD
        
        self.assertFloatEqual(\
        self.q1._get_three_of_a_kind(3), 0.22222222222222221)  # - D|ABC
        
        self.assertEqual(self.q2._get_three_of_a_kind(0), 0)   # - A|BCD
        
        self.assertFloatEqual(\
        self.q2._get_three_of_a_kind(1), 0.22222222222222221)   # - B|ACD
        
        self.assertEqual(self.q2._get_three_of_a_kind(2), 0)   # - C|ABD
        self.assertFloatEqual(\
        self.q2._get_three_of_a_kind(3), 0.1111111111111111)   # - D|ABC
        
        self.assertFloatEqual(\
        self.q3._get_three_of_a_kind(0), 0.1111111111111111)   # - A|BCD
        
        self.assertEqual(self.q3._get_three_of_a_kind(1), 0)   # - B|ACD
        self.assertEqual(self.q3._get_three_of_a_kind(2), 0)   # - C|ABD
        self.assertEqual(self.q3._get_three_of_a_kind(3), 0)   # - D|ABC
        
    def test_get_pairwise_cong(self):
        """Quartet._get_pairwise_cong: should return % of cols with 2 pairs.
        """
        self.assertEqual(self.q1._get_pairwise_cong((0, 1)), 0) # - 'AB'
        self.assertFloatEqual(\
        self.q1._get_pairwise_cong((1, 2)), 0.1111111111111111) # - 'AD'
        
        self.assertEqual(self.q1._get_pairwise_cong((1, 3)), 0) # - 'AC'
        self.assertFloatEqual(\
        self.q2._get_pairwise_cong((0, 1)), 0.3333333333333333) # - 'AB'
        
        self.assertFloatEqual(\
        self.q2._get_pairwise_cong((1, 2)), 0.1111111111111111) # - 'AD'
        
        self.assertFloatEqual(\
        self.q2._get_pairwise_cong((1, 3)), 0.1111111111111111) # - 'AC'
        
        self.assertFloatEqual(\
        self.q3._get_pairwise_cong((0, 1)), 0.22222222222222221)# - 'AB'
        
        self.assertFloatEqual(\
        self.q3._get_pairwise_cong((1, 2)), 0.4444444444444444) # - 'AD'
    
        self.assertFloatEqual(\
        self.q3._get_pairwise_cong((1, 3)), 0.1111111111111111) # - 'AC'
        
    def test_analyze_alpha(self):
        """Quartet.analyze: should calculate A|BCD correctly for valid input.
        """
        self.assertEqual(self.q1.analyze()['A|BCD'], 0)
        self.assertEqual(self.q2.analyze()['A|BCD'], 0)
        self.assertFloatEqual(\
        self.q3.analyze()['A|BCD'], 0.1111111111111111)
        
    def test_analyze_beta(self):
        """Quartet.analyze: should calculate B|ACD correctly for valid input.
        """
        self.assertFloatEqual(\
        self.q1.analyze()['B|ACD'], 0.11111111111110)
        
        self.assertFloatEqual(\
        self.q2.analyze()['B|ACD'], 0.22222222222222221)
        
        self.assertEqual(self.q3.analyze()['B|ACD'], 0)
        
    def test_analyze_gamma(self):
        """Quartet.analyze: should calculate C|ABD correctly for valid input.
        """
        self.assertFloatEqual(\
        self.q1.analyze()['C|ABD'], 0.3333333333333333)
        
        self.assertEqual(self.q2.analyze()['C|ABD'], 0)
        self.assertEqual(self.q3.analyze()['C|ABD'], 0)
        
    def test_analyze_delta(self):
        """Quartet.analyze: should calculate D|ABC correctly for valid input.
        """
        self.assertFloatEqual(\
        self.q1.analyze()['D|ABC'], 0.22222222222222221)
        
        self.assertFloatEqual(\
        self.q2.analyze()['D|ABC'], 0.1111111111111111)
        
        self.assertEqual(self.q3.analyze()['D|ABC'], 0)
        
    def test_analyze_AB(self):
        """Quartet.analyze: should calculate AB correctly.
        """
        self.assertFloatEqual(self.q1.analyze()['AB|CD'], 0/9)
        self.assertFloatEqual(self.q2.analyze()['AB|CD'], 3/9)        
        self.assertFloatEqual(self.q3.analyze()['AB|CD'], 2/9)
        
    def test_analyze_BC(self):
        """Quartet.analyze: should calculate AD correctly.
        """
        self.assertFloatEqual(self.q1.analyze()['AD|BC'], 1/9)
        self.assertFloatEqual(self.q2.analyze()['AD|BC'], 1/9)
        self.assertFloatEqual(self.q3.analyze()['AD|BC'], 4/9)

    def test_analyze_AC(self):
        """Quartet.analyze: should calculate AC correctly.
        """
        self.assertFloatEqual(self.q1.analyze()['AC|BD'], 0/9)
        self.assertFloatEqual(self.q2.analyze()['AC|BD'], 1/9)
        self.assertFloatEqual(self.q3.analyze()['AC|BD'], 1/9)
        
    def test_analyze_d0(self):
        """Quartet.analyze: should calculate d0 correctly.
        """
        self.assertFloatEqual(self.q1.analyze()['d0'], 2/9)
        self.assertFloatEqual(self.q2.analyze()['d0'], 1/9)
        self.assertFloatEqual(self.q3.analyze()['d0'], 1/9)
        
    def test_analyze_d1(self):
        """Quartet.analyze: should calculate d1 correctly.
        """
        self.assertFloatEqual(self.q1.analyze()['d1'], 6/9)
        self.assertFloatEqual(self.q2.analyze()['d1'], 3/9)
        self.assertFloatEqual(self.q3.analyze()['d1'], 1/9)

    def test_analyze_d2(self):
        """Quartet.analyze: should calculate d1 correctly.
        """
        self.assertFloatEqual(self.q1.analyze()['d2'], 1/9)
        self.assertFloatEqual(self.q2.analyze()['d2'], 5/9)
        self.assertFloatEqual(self.q3.analyze()['d2'], 7/9)
        
    def test_degap(self):
        """Quartet.degap: should remove gaps from self.
        """
        qq = self.q_gaps
        self.assertFloatEqual(qq.QuartetLen, 9.0)
        
        # Let's remove two gaps!
        qq.degap()
        self.assertFloatEqual(qq.QuartetLen, 4.0)
        
class BinaryQuartetTests(TestCaseBugFix):
    """Tests for BinaryQuartet object.
    """
    def setUp(self):
        """Creates environment for all tests of BinaryQuartet object.
        """
        b1 = ByteSequence('110010101', Name='b1')
        b2 = ByteSequence('110010000', Name='b2')
        b3 = ByteSequence('010100001', Name='b3')
        b4 = ByteSequence('110011111', Name='b4')
        b5 = ByteSequence('111111110', Name='b5')
        b6 = ByteSequence('011000010', Name='b6')
        
        self.four_byte_seqs_DIGITS = [b1, b2, b3, b4]
        self.REF_four_byte_seqs_DIGITS = \
array([['1', '1', '0', '1'],
       ['1', '1', '1', '1'],
       ['0', '0', '0', '0'],
       ['0', '0', '1', '0'],
       ['1', '1', '0', '1'],
       ['0', '0', '0', '1'],
       ['1', '0', '0', '1'],
       ['0', '0', '0', '1'],
       ['1', '0', '1', '1']], 
      dtype='|S1')
        
        self.valid_quartet_DIGITS = DenseAlignment(self.four_byte_seqs_DIGITS)
        self.invalid_quartet_DIGITS = DenseAlignment([b1, b2, b3, b4, b5])
        
        self.q1_BINARY = BinaryQuartet(self.valid_quartet_DIGITS)
        self.q2_BINARY = BinaryQuartet(DenseAlignment([b2,b3,b4,b5]))
        self.q3_BINARY = BinaryQuartet(DenseAlignment([b3,b4,b5,b6]))

        self.q_gaps = BinaryQuartet(DenseAlignment([b5, b6, b4,\
        ByteSequence('0110-----', Name='gaps')]))
        
    ## THIS SINGLE TEST BELOW CHECKS WHAT HAPPENS IF INPUT IS INVALID
    def test_init(self):
        """BinaryQuartet.__init__: should set self.Quartet & raise QuartetError
        """
        #
        BS = ByteSequence
        RS = RnaSequence
        DA = DenseAlignment
        #
        
        # We test here what happens if input is correct ...
        
        self.assertEqualItems(self.q1_BINARY.Quartet.tolist(),\
        self.REF_four_byte_seqs_DIGITS.tolist())
        
        # ... and whether QuartetError is raised when it is incorrect.
        self.assertRaises(\
        QuartetError, BinaryQuartet, self.invalid_quartet_DIGITS)
        
        self.assertRaises(QuartetError, BinaryQuartet, BS('111'))
        
        self.assertRaises(\
        QuartetError, BinaryQuartet, [BS('1'), BS('1'), BS('1'), BS('1')])
        
        self.assertRaises(QuartetError, BinaryQuartet, 'AUGC')
        
        self.assertRaises(\
        QuartetError, BinaryQuartet, [Alignment('U'), Alignment('C'),\
        Alignment('G'), Alignment('A')])
        
        self.assertRaises(\
        QuartetError, BinaryQuartet, DA([BS('1'), BS('1'), BS('1')]))
        
        self.assertRaises(QuartetError, BinaryQuartet, DA([BS('1'), BS('1')]))
        
        self.assertRaises(\
        QuartetError, BinaryQuartet, self.invalid_quartet_DIGITS)
        
        self.assertRaises(QuartetError, BinaryQuartet, BS('111'))
        
        self.assertRaises(\
        QuartetError, BinaryQuartet, [BS('1'), BS('1'), BS('1'), BS('1')])
        
        self.assertRaises(QuartetError, BinaryQuartet, 'ACUGT')
        
        self.assertRaises(QuartetError, BinaryQuartet, Alignment('ACUGT'))
        self.assertRaises(QuartetError, BinaryQuartet, Alignment('-----'))
        
        self.assertRaises(QuartetError, BinaryQuartet,\
        [Alignment('U'), Alignment('C'), Alignment('G'), Alignment('A')])
        
        self.assertRaises(QuartetError, BinaryQuartet, \
        [Alignment('B'), Alignment('B'), Alignment('-'), Alignment('A')])
        
    def test_str(self):
        """BinaryQuartet.__str__: should return string repre. of BinaryQuartet
        """
        exp = str(self.REF_four_byte_seqs_DIGITS)       # DenseAlignment object
        obs = str(self.q1_BINARY)                       # BinaryQuartet object 
        
        self.assertEqual(obs, exp)  # Although the objects are different, the
                                    # output should be the same, because
                                    # self.q1_BINARY was initialized with
                                    # self.valid_quartet_DIGITS.
        
    def test_get_all_pos_equal(self):
        """BinaryQuartet._get_all_pos_equal: should return % of cols of == pos
        """
        qb = self.q1_BINARY
        qb.analyze()
        self.assertAlmostEqual(qb._get_all_pos_equal(), 0.222222, places=6)
        
    def test_get_d3_d4_binary_space(self):
        """BinaryQuartet._get_d3_d4 should return 0, 0 in case of binary space
        """
        d3, d4 = self.q1_BINARY._get_d3_d4()
        self.assertFloatEqual(d3, 0)
        self.assertFloatEqual(d4, 0)
        self.assertFloatEqual(self.q1_BINARY.Params['AB|C|D'], 0)
        self.assertFloatEqual(self.q1_BINARY.Params['AC|B|D'], 0)
        self.assertFloatEqual(self.q1_BINARY.Params['AD|B|C'], 0)
        self.assertFloatEqual(self.q1_BINARY.Params['BC|A|D'], 0)
        self.assertFloatEqual(self.q1_BINARY.Params['BD|A|C'], 0)
        self.assertFloatEqual(self.q1_BINARY.Params['CD|A|B'], 0)
        
    def test_get_three_of_a_kind(self):
        """BinaryQuartet._get_three_of_a_kind: returns cols' % with 3 eq pos.
        """
        self.assertEqual(self.q1_BINARY._get_three_of_a_kind(0), 0)   # - A|BCD
        
        self.assertFloatEqual(\
        self.q1_BINARY._get_three_of_a_kind(1), 0.1111111111111111)   # - B|ACD
        
        self.assertFloatEqual(\
        self.q1_BINARY._get_three_of_a_kind(2), 0.3333333333333333)   # - C|ABD
        
        self.assertFloatEqual(\
        self.q1_BINARY._get_three_of_a_kind(3), 0.22222222222222221)  # - D|ABC
        
        self.assertEqual(self.q2_BINARY._get_three_of_a_kind(0), 0)   # - A|BCD
        
        self.assertFloatEqual(\
        self.q2_BINARY._get_three_of_a_kind(1), 0.22222222222222221)   # - B|ACD
        
        self.assertEqual(self.q2_BINARY._get_three_of_a_kind(2), 0)   # - C|ABD
        self.assertFloatEqual(\
        self.q2_BINARY._get_three_of_a_kind(3), 0.1111111111111111)   # - D|ABC
        
        self.assertFloatEqual(\
        self.q3_BINARY._get_three_of_a_kind(0), 0.1111111111111111)   # - A|BCD
        
        self.assertEqual(self.q3_BINARY._get_three_of_a_kind(1), 0)   # - B|ACD
        self.assertEqual(self.q3_BINARY._get_three_of_a_kind(2), 0)   # - C|ABD
        self.assertEqual(self.q3_BINARY._get_three_of_a_kind(3), 0)   # - D|ABC
        
    def test_get_pairwise_cong(self):
        """BinaryQuartet._get_pairwise_cong: should return % of cols with2pairs
        """
        self.assertEqual(self.q1_BINARY._get_pairwise_cong((0, 1)), 0) # - 'AB'
        self.assertFloatEqual(\
        self.q1_BINARY._get_pairwise_cong((1, 2)), 0.1111111111111111) # - 'AD'
        
        self.assertEqual(self.q1_BINARY._get_pairwise_cong((1, 3)), 0) # - 'AC'
        self.assertFloatEqual(\
        self.q2_BINARY._get_pairwise_cong((0, 1)), 0.3333333333333333) # - 'AB'
        
        self.assertFloatEqual(\
        self.q2_BINARY._get_pairwise_cong((1, 2)), 0.1111111111111111) # - 'AD'
        
        self.assertFloatEqual(\
        self.q2_BINARY._get_pairwise_cong((1, 3)), 0.1111111111111111) # - 'AC'
        
        self.assertFloatEqual(\
        self.q3_BINARY._get_pairwise_cong((0, 1)), 0.22222222222222221)# - 'AB'
        
        self.assertFloatEqual(\
        self.q3_BINARY._get_pairwise_cong((1, 2)), 0.4444444444444444) # - 'AD'
    
        self.assertFloatEqual(\
        self.q3_BINARY._get_pairwise_cong((1, 3)), 0.1111111111111111) # - 'AC'
        
    def test_analyze_alpha(self):
        """BinaryQuartet.analyze: calculates A|BCD correctly for valid input
        """
        self.assertEqual(self.q1_BINARY.analyze()['A|BCD'], 0)
        self.assertEqual(self.q2_BINARY.analyze()['A|BCD'], 0)
        self.assertFloatEqual(\
        self.q3_BINARY.analyze()['A|BCD'], 0.1111111111111111)
        
    def test_analyze_beta(self):
        """BinaryQuartet.analyze: calculates B|ACD correctly for valid input
        """
        self.assertFloatEqual(\
        self.q1_BINARY.analyze()['B|ACD'], 0.11111111111110)
        
        self.assertFloatEqual(\
        self.q2_BINARY.analyze()['B|ACD'], 0.22222222222222221)
        
        self.assertEqual(self.q3_BINARY.analyze()['B|ACD'], 0)
        
    def test_analyze_gamma(self):
        """BinaryQuartet.analyze: calculates C|ABD correctly for valid input
        """
        self.assertFloatEqual(\
        self.q1_BINARY.analyze()['C|ABD'], 0.3333333333333333)
        
        self.assertEqual(self.q2_BINARY.analyze()['C|ABD'], 0)
        self.assertEqual(self.q3_BINARY.analyze()['C|ABD'], 0)
        
    def test_analyze_delta(self):
        """BinaryQuartet.analyze: calculates D|ABC correctly for valid input
        """
        self.assertFloatEqual(\
        self.q1_BINARY.analyze()['D|ABC'], 0.22222222222222221)
        
        self.assertFloatEqual(\
        self.q2_BINARY.analyze()['D|ABC'], 0.1111111111111111)
        
        self.assertEqual(self.q3_BINARY.analyze()['D|ABC'], 0)
        
    def test_analyze_AB(self):
        """BinaryQuartet.analyze: should calculate AB correctly.
        """
        self.assertFloatEqual(self.q1_BINARY.analyze()['AB|CD'], 0/9)
        self.assertFloatEqual(self.q2_BINARY.analyze()['AB|CD'], 3/9)        
        self.assertFloatEqual(self.q3_BINARY.analyze()['AB|CD'], 2/9)
        
    def test_analyze_BC(self):
        """BinaryQuartet.analyze: should calculate AD correctly.
        """
        self.assertFloatEqual(self.q1_BINARY.analyze()['AD|BC'], 1/9)
        self.assertFloatEqual(self.q2_BINARY.analyze()['AD|BC'], 1/9)
        self.assertFloatEqual(self.q3_BINARY.analyze()['AD|BC'], 4/9)

    def test_analyze_AC(self):
        """BinaryQuartet.analyze: should calculate AC correctly.
        """
        self.assertFloatEqual(self.q1_BINARY.analyze()['AC|BD'], 0/9)
        self.assertFloatEqual(self.q2_BINARY.analyze()['AC|BD'], 1/9)
        self.assertFloatEqual(self.q3_BINARY.analyze()['AC|BD'], 1/9)
        
    def test_analyze_d0(self):
        """BinaryQuartet.analyze: should calculate d0 correctly.
        """
        self.assertFloatEqual(self.q1_BINARY.analyze()['d0'], 2/9)
        self.assertFloatEqual(self.q2_BINARY.analyze()['d0'], 1/9)
        self.assertFloatEqual(self.q3_BINARY.analyze()['d0'], 1/9)
        
    def test_analyze_d1(self):
        """BinaryQuartet.analyze: should calculate d1 correctly.
        """
        self.assertFloatEqual(self.q1_BINARY.analyze()['d1'], 6/9)
        self.assertFloatEqual(self.q2_BINARY.analyze()['d1'], 3/9)
        self.assertFloatEqual(self.q3_BINARY.analyze()['d1'], 1/9)

    def test_analyze_d2(self):
        """BinaryQuartet.analyze: should calculate d1 correctly.
        """
        self.assertFloatEqual(self.q1_BINARY.analyze()['d2'], 1/9)
        self.assertFloatEqual(self.q2_BINARY.analyze()['d2'], 5/9)
        self.assertFloatEqual(self.q3_BINARY.analyze()['d2'], 7/9)
        
    def test_degap(self):
        """BinaryQuartet.degap: should remove gaps from self.
        """
        qq = self.q_gaps
        self.assertFloatEqual(qq.QuartetLen, 9.0)
        
        # Let's remove two gaps!
        qq.degap()
        self.assertFloatEqual(qq.QuartetLen, 4.0)

class QuaternaryQuartetTests(TestCaseBugFix):
    """Tests for QuaternaryQuartet object.
    """
    def setUp(self):
        """Creates environment for all tests of QuaternaryQuartet object.
        """
        b1 = ByteSequence('112233441123241214', Name='b1')
        b2 = ByteSequence('112233111141124121', Name='b2')
        b3 = ByteSequence('444444112223112333', Name='b3')
        b4 = ByteSequence('444433111312424342', Name='b4')
        b5 = ByteSequence('212233141111111111', Name='b5')
        b6 = ByteSequence('232233441444422231', Name='b6')
        
        self.four_byte_seqs_DIGITS = [b1, b2, b3, b4]
        self.REF_four_byte_seqs_DIGITS = \
array([['1', '1', '4', '4'],
       ['1', '1', '4', '4'],
       ['2', '2', '4', '4'],
       ['2', '2', '4', '4'],
       ['3', '3', '4', '3'],
       ['3', '3', '4', '3'],
       ['4', '1', '1', '1'],
       ['4', '1', '1', '1'],
       ['1', '1', '2', '1'],
       ['1', '1', '2', '3'],
       ['2', '4', '2', '1'],
       ['3', '1', '3', '2'],
       ['2', '1', '1', '4'],
       ['4', '2', '1', '2'],
       ['1', '4', '2', '4'],
       ['2', '1', '3', '3'],
       ['1', '2', '3', '4'],
       ['4', '1', '3', '2']],
      dtype='|S1')
    
        b7 = ByteSequence('12344', Name='b7')
        b8 = ByteSequence('14412', Name='b8')
        b9 = ByteSequence('32111', Name='b9')
        b10 = ByteSequence('44444', Name='b10')
        b11 = ByteSequence('--412', Name='b11')
        
        self.valid_quartet_DIGITS = DenseAlignment(self.four_byte_seqs_DIGITS)
        self.invalid_quartet_DIGITS = DenseAlignment([b1, b2, b3, b4, b5])
        
        self.q1 = QuaternaryQuartet(self.valid_quartet_DIGITS)
        self.q2 = QuaternaryQuartet(DenseAlignment([b2,b3,b4,b5]))
        self.q3 = QuaternaryQuartet(DenseAlignment([b3,b4,b5,b6]))
        self.q5 = QuaternaryQuartet(DenseAlignment([b7,b8,b9,b11]))
        
    ## THIS SINGLE TEST BELOW CHECKS WHAT HAPPENS IF INPUT IS INVALID
    def test_init(self):
        """QuaternaryQuartet.__init__: sets self.Quartet & raises QuartetError
        """
        #
        BS = ByteSequence
        RS = RnaSequence
        DA = DenseAlignment
        #
        
        # We test here what happens if input is correct ...
        
        self.assertEqualItems(self.q1.Quartet.tolist(),\
            self.REF_four_byte_seqs_DIGITS.tolist())
        
        # ... and whether QuartetError is raised when it is incorrect.
        self.assertRaises(\
            QuartetError, QuaternaryQuartet, self.invalid_quartet_DIGITS)
        
        self.assertRaises(QuartetError, QuaternaryQuartet, BS('111'))
        
        self.assertRaises(\
        QuartetError, QuaternaryQuartet, [BS('1'), BS('1'), BS('1'), BS('1')])
        
        self.assertRaises(QuartetError, QuaternaryQuartet, 'AUGC')
        
        self.assertRaises(\
        QuartetError, QuaternaryQuartet, [Alignment('U'), Alignment('C'),\
        Alignment('G'), Alignment('A')])
        
        self.assertRaises(\
        QuartetError, QuaternaryQuartet, DA([BS('1'), BS('1'), BS('1')]))
        
        self.assertRaises(QuartetError, QuaternaryQuartet, DA([BS('1'), BS('1')]))
        
        self.assertRaises(\
        QuartetError, QuaternaryQuartet, self.invalid_quartet_DIGITS)
        
        self.assertRaises(QuartetError, QuaternaryQuartet, BS('111'))
        
        self.assertRaises(\
        QuartetError, QuaternaryQuartet, [BS('1'), BS('1'), BS('1'), BS('1')])
        
        self.assertRaises(QuartetError, QuaternaryQuartet, 'ACUGT')
        
        self.assertRaises(QuartetError, QuaternaryQuartet, Alignment('ACUGT'))
        self.assertRaises(QuartetError, QuaternaryQuartet, Alignment('-----'))
        
        self.assertRaises(QuartetError, QuaternaryQuartet,\
        [Alignment('U'), Alignment('C'), Alignment('G'), Alignment('A')])
        
        self.assertRaises(QuartetError, QuaternaryQuartet, \
        [Alignment('B'), Alignment('B'), Alignment('-'), Alignment('A')])
        
    def test_str(self):
        """QuaternaryQuartet.__str__: returns str repr. of QuaternaryQuartet
        """
        exp = str(self.REF_four_byte_seqs_DIGITS)       # DenseAlignment object
        obs = str(self.q1)                       # QuaternaryQuartet object 
        
        self.assertEqual(obs, exp)  # Although the objects are different, the
                                    # output should be the same, because
                                    # self.q1 was initialized with
                                    # self.valid_quartet_DIGITS.
                                    
    def test_get_d3_d4_quaternary_space(self):
        """QuaternaryQuartet._get_d3_d4 should return d3 & d4 > 0
        """
        self.assertEqual(self.q1.QuartetLen, 18)
        d3, d4 = self.q1._get_d3_d4()
        self.assertAlmostEqual(d3, 7/18)
        self.assertEqual(d4, 2/18)
        
        self.assertFloatEqual(self.q1.Params['AB|C|D'], 1/18)
        self.assertFloatEqual(self.q1.Params['AC|B|D'], 2/18)
        self.assertFloatEqual(self.q1.Params['AD|B|C'], 0/18)
        self.assertFloatEqual(self.q1.Params['BC|A|D'], 1/18)
        self.assertFloatEqual(self.q1.Params['BD|A|C'], 2/18)
        self.assertFloatEqual(self.q1.Params['CD|A|B'], 1/18)
        
    def test_get_all_pos_equal(self):
        """QuaternaryQuartet._get_all_pos_equal: returns % of cols of == pos
        """
        self.q1.analyze()
        self.q2.analyze()
        self.q3.analyze()
        
        self.assertFloatEqual(self.q1._get_all_pos_equal(), 0.0)
        self.assertFloatEqual(self.q2._get_all_pos_equal(), 1/18)
        self.assertFloatEqual(self.q3._get_all_pos_equal(), 0)
        
    def test_get_three_of_a_kind(self):
        """QuaternaryQuartet._get_three_of_a_kind: ReturnsCols'%With3EqPos
        """
        # self.q1
        self.assertFloatEqual(\
        self.q1._get_three_of_a_kind(0), 2/18)             # - A|BCD
        
        self.assertFloatEqual(\
        self.q1._get_three_of_a_kind(1), 0)                  # - B|ACD
        
        self.assertFloatEqual(\
        self.q1._get_three_of_a_kind(2), 3/18)             # - C|ABD
        
        self.assertFloatEqual(\
        self.q1._get_three_of_a_kind(3), 0)                  # - D|ABC

        # self.q2
        self.assertFloatEqual(\
        self.q2._get_three_of_a_kind(0), 0)                    # - A|BCD
        
        self.assertFloatEqual(\
        self.q2._get_three_of_a_kind(1), 3/18)             # - B|ACD
        
        self.assertFloatEqual(\
        self.q2._get_three_of_a_kind(2), 1/18)             # - C|ABD
        
        self.assertFloatEqual(\
        self.q2._get_three_of_a_kind(3), 1/18)             # - D|ABC

        # self.q3
        self.assertFloatEqual(\
        self.q3._get_three_of_a_kind(0), 3/18)             # - A|BCD
        
        self.assertFloatEqual(\
        self.q3._get_three_of_a_kind(1), 0)                  # - B|ACD
        
        self.assertFloatEqual(\
        self.q3._get_three_of_a_kind(2), 0)                  # - C|ABD
        
        self.assertFloatEqual(\
        self.q3._get_three_of_a_kind(3), 1/18)             # - D|ABC
        
    def test_get_pairwise_cong(self):
        """QuaternaryQuartet._get_pairwise_cong: returns % of cols with2pairs.
        """
        # self.q1
        self.assertFloatEqual(\
        self.q1._get_pairwise_cong((0, 1)), 4/18)           # - 'AB'
        
        self.assertFloatEqual(\
        self.q1._get_pairwise_cong((1, 2)), 0)                # - 'AD'
        
        self.assertFloatEqual(\
        self.q1._get_pairwise_cong((1, 3)), 0)                # - 'AC'

        # self.q2
        self.assertFloatEqual(\
        self.q2._get_pairwise_cong((0, 1)), 0)                # - 'AB'
        
        self.assertFloatEqual(\
        self.q2._get_pairwise_cong((1, 2)), 4/18)           # - 'AD'
        
        self.assertFloatEqual(\
        self.q2._get_pairwise_cong((1, 3)), 1/18)           # - 'AC'

        # self.q3
        self.assertFloatEqual(\
        self.q3._get_pairwise_cong((0, 1)), 4/18)           # - 'AB'
        
        self.assertFloatEqual(\
        self.q3._get_pairwise_cong((1, 2)), 0)                # - 'AD'
    
        self.assertFloatEqual(\
        self.q3._get_pairwise_cong((1, 3)), 2/18)           # - 'AC'
        
    def test_analyze_alpha(self):
        """QuaternaryQuartet.analyze:calculates A|BCD correctly for valid input
        """
        self.assertFloatEqual(\
        self.q1.analyze()['A|BCD'], 2/18)
        
        self.assertFloatEqual(\
        self.q2.analyze()['A|BCD'], 0)
        
        self.assertFloatEqual(\
        self.q3.analyze()['A|BCD'], 3/18)
        
    def test_analyze_beta(self):
        """QuaternaryQuartet.analyze: calculates B|ACD correctly for valid input
        """
        self.assertFloatEqual(\
        self.q1.analyze()['B|ACD'], 0)
        
        self.assertFloatEqual(\
        self.q2.analyze()['B|ACD'], 3/18)
        
        self.assertFloatEqual(\
        self.q3.analyze()['B|ACD'], 0)
        
    def test_analyze_gamma(self):
        """QuaternaryQuartet.analyze:calculates C|ABD correctly for valid input
        """
        self.assertFloatEqual(\
        self.q1.analyze()['C|ABD'], 3/18)
        
        self.assertFloatEqual(\
        self.q2.analyze()['C|ABD'], 1/18)
        
        self.assertFloatEqual(\
        self.q3.analyze()['C|ABD'], 0)
        
    def test_analyze_delta(self):
        """QuaternaryQuartet.analyze:calculates D|ABC correctly for valid input
        """
        self.assertFloatEqual(\
        self.q1.analyze()['D|ABC'], 0)
        
        self.assertFloatEqual(\
        self.q2.analyze()['D|ABC'], 1/18)
        
        self.assertFloatEqual(\
        self.q3.analyze()['D|ABC'], 1/18)
        
    def test_analyze_AB(self):
        """QuaternaryQuartet.analyze: should calculate AB correctly.
        """
        self.assertFloatEqual(self.q1.analyze()['AB|CD'], 4/18)
        self.assertFloatEqual(self.q2.analyze()['AB|CD'], 0)
        self.assertFloatEqual(self.q3.analyze()['AB|CD'], 4/18)
        
    def test_analyze_BC(self):
        """QuaternaryQuartet.analyze: should calculate AD correctly.
        """
        self.assertFloatEqual(self.q1.analyze()['AD|BC'], 0)
        self.assertFloatEqual(self.q2.analyze()['AD|BC'], 4/18)
        self.assertFloatEqual(self.q3.analyze()['AD|BC'], 0)

    def test_analyze_AC(self):
        """QuaternaryQuartet.analyze: should calculate AC correctly.
        """
        self.assertFloatEqual(self.q1.analyze()['AC|BD'], 0)
        self.assertFloatEqual(self.q2.analyze()['AC|BD'], 1/18)
        self.assertFloatEqual(self.q3.analyze()['AC|BD'], 2/18)

    def test_analyze_d0(self):
        """QuaternaryQuartet.analyze: should calculate d0 correctly.
        """
        self.assertFloatEqual(self.q1.analyze()['d0'], 0/18)
        self.assertFloatEqual(self.q2.analyze()['d0'], 1/18)
        self.assertFloatEqual(self.q3.analyze()['d0'], 0/18)

    def test_analyze_d1(self):
        """QuaternaryQuartet.analyze: should calculate d1 correctly.
        """
        self.assertFloatEqual(self.q1.analyze()['d1'], 5/18)
        self.assertFloatEqual(self.q2.analyze()['d1'], 5/18)
        self.assertFloatEqual(self.q3.analyze()['d1'], 4/18)

    def test_analyze_d2(self):
        """QuaternaryQuartet.analyze: should calculate d2 correctly.
        """
        self.assertFloatEqual(self.q1.analyze()['d2'], 4/18)
        self.assertFloatEqual(self.q2.analyze()['d2'], 5/18)
        self.assertFloatEqual(self.q3.analyze()['d2'], 6/18)

    def test_analyze_d3(self):
        """QuaternaryQuartet.analyze: should calculate d3 correctly.
        """
        self.assertFloatEqual(self.q1.analyze()['d3'], 7/18)
        self.assertFloatEqual(self.q2.analyze()['d3'], 6/18)
        self.assertFloatEqual(self.q3.analyze()['d3'], 6/18)

    def test_analyze_d4(self):
        """QuaternaryQuartet.analyze: should calculate d4 correctly.
        """
        self.assertFloatEqual(self.q1.analyze()['d4'], 2/18)
        self.assertFloatEqual(self.q2.analyze()['d4'], 1/18)
        self.assertFloatEqual(self.q3.analyze()['d4'], 2/18)
        
    def test_degap(self):
        """QuaternaryQuartet.degap: should remove gaps from self.
        """
        self.q5.degap()
        self.assertFloatEqual(self.q5.QuartetLen, 3.0)
        
class QuarterGeneratorSetUp(TestCaseBugFix):
    """Data for tests of all functions cooperating with quartet_generator.
    """
    def setUp(self):
        """Creates all the data sets required by quartet_generator and co.
        """
        b1 = ByteSequence('110010101', Name='b1')
        b2 = ByteSequence('110010000', Name='b2')
        b3 = ByteSequence('010100001', Name='b3')
        b4 = ByteSequence('110011111', Name='b4')
        b5 = ByteSequence('111111110', Name='b5')
        
        self.da_DIGITS = DenseAlignment([b1, b2, b3, b4, b5])
        self.reference_da_DIGITS = [
array([['1', '1', '0', '1'],
       ['1', '1', '1', '1'],
       ['0', '0', '0', '0'],
       ['0', '0', '1', '0'],
       ['1', '1', '0', '1'],
       ['0', '0', '0', '1'],
       ['1', '0', '0', '1'],
       ['0', '0', '0', '1'],
       ['1', '0', '1', '1']], 
      dtype='|S1'),
array([['1', '1', '0', '1'],
       ['1', '1', '1', '1'],
       ['0', '0', '0', '1'],
       ['0', '0', '1', '1'],
       ['1', '1', '0', '1'],
       ['0', '0', '0', '1'],
       ['1', '0', '0', '1'],
       ['0', '0', '0', '1'],
       ['1', '0', '1', '0']], 
      dtype='|S1'),
array([['1', '1', '1', '1'],
       ['1', '1', '1', '1'],
       ['0', '0', '0', '1'],
       ['0', '0', '0', '1'],
       ['1', '1', '1', '1'],
       ['0', '0', '1', '1'],
       ['1', '0', '1', '1'],
       ['0', '0', '1', '1'],
       ['1', '0', '1', '0']], 
      dtype='|S1'),
array([['1', '0', '1', '1'],
       ['1', '1', '1', '1'],
       ['0', '0', '0', '1'],
       ['0', '1', '0', '1'],
       ['1', '0', '1', '1'],
       ['0', '0', '1', '1'],
       ['1', '0', '1', '1'],
       ['0', '0', '1', '1'],
       ['1', '1', '1', '0']], 
      dtype='|S1'),
array([['1', '0', '1', '1'],
       ['1', '1', '1', '1'],
       ['0', '0', '0', '1'],
       ['0', '1', '0', '1'],
       ['1', '0', '1', '1'],
       ['0', '0', '1', '1'],
       ['0', '0', '1', '1'],
       ['0', '0', '1', '1'],
       ['0', '1', '1', '0']], 
      dtype='|S1')]
        
        b6 = ByteSequence('_-+0-)(%^', Name='b6')
        b7 = ByteSequence('_-+00)(%^', Name='b7')
        b8 = ByteSequence('_________', Name='b8')
        b9 = ByteSequence('-----)(%^', Name='b9')
        b10 = ByteSequence('00-00)(%-', Name='b10')
        b11 = ByteSequence('_0-0-----', Name='b11')
        b15 = ByteSequence('_0-------', Name='b15')
        b16 = ByteSequence('_000-----', Name='b16')
        b17 = ByteSequence('_0000___-', Name='b17')
        b18 = ByteSequence('_0000____', Name='b18')
        
        self.da_CHARS = DenseAlignment([b7, b8, b9, b10, b11])
        
        b12 = ByteSequence('_0_0')
        b13 = ByteSequence('0000')
        b14 = ByteSequence('____')
        
        self.da1 = DenseAlignment([b15, b16, b17])
        self.da2 = DenseAlignment([b9, b10, b11])
        self.da3 = DenseAlignment([b6, b7, b8])
        self.da4 = DenseAlignment([b3, b4, b5])
        self.da5 = DenseAlignment([b10, b11, b18])
        
        self.quartets_from_4_das = [
        DenseAlignment([b15, b9, b6, b3]),
        DenseAlignment([b15, b9, b6, b4]),
        DenseAlignment([b15, b9, b6, b5]),
        DenseAlignment([b15, b9, b7, b3]),
        DenseAlignment([b15, b9, b7, b4]),
        DenseAlignment([b15, b9, b7, b5]),
        DenseAlignment([b15, b9, b8, b3]),
        DenseAlignment([b15, b9, b8, b4]),
        DenseAlignment([b15, b9, b8, b5]),      # 9th
        DenseAlignment([b15, b10, b6, b3]),
        DenseAlignment([b15, b10, b6, b4]),
        DenseAlignment([b15, b10, b6, b5]),
        DenseAlignment([b15, b10, b7, b3]),
        DenseAlignment([b15, b10, b7, b4]),
        DenseAlignment([b15, b10, b7, b5]),
        DenseAlignment([b15, b10, b8, b3]),
        DenseAlignment([b15, b10, b8, b4]),
        DenseAlignment([b15, b10, b8, b5]),     # 18th
        DenseAlignment([b15, b11, b6, b3]),
        DenseAlignment([b15, b11, b6, b4]),
        DenseAlignment([b15, b11, b6, b5]),
        DenseAlignment([b15, b11, b7, b3]),
        DenseAlignment([b15, b11, b7, b4]),
        DenseAlignment([b15, b11, b7, b5]),
        DenseAlignment([b15, b11, b8, b3]),
        DenseAlignment([b15, b11, b8, b4]),
        DenseAlignment([b15, b11, b8, b5]),     # 27th
        DenseAlignment([b16, b9, b6, b3]),
        DenseAlignment([b16, b9, b6, b4]),
        DenseAlignment([b16, b9, b6, b5]),
        DenseAlignment([b16, b9, b7, b3]),
        DenseAlignment([b16, b9, b7, b4]),
        DenseAlignment([b16, b9, b7, b5]),
        DenseAlignment([b16, b9, b8, b3]),
        DenseAlignment([b16, b9, b8, b4]),
        DenseAlignment([b16, b9, b8, b5]),      # 36th
        DenseAlignment([b16, b10, b6, b3]),
        DenseAlignment([b16, b10, b6, b4]),
        DenseAlignment([b16, b10, b6, b5]),
        DenseAlignment([b16, b10, b7, b3]),
        DenseAlignment([b16, b10, b7, b4]),
        DenseAlignment([b16, b10, b7, b5]),
        DenseAlignment([b16, b10, b8, b3]),
        DenseAlignment([b16, b10, b8, b4]),
        DenseAlignment([b16, b10, b8, b5]),     # 45th
        DenseAlignment([b16, b11, b6, b3]),
        DenseAlignment([b16, b11, b6, b4]),
        DenseAlignment([b16, b11, b6, b5]),
        DenseAlignment([b16, b11, b7, b3]),
        DenseAlignment([b16, b11, b7, b4]),
        DenseAlignment([b16, b11, b7, b5]),
        DenseAlignment([b16, b11, b8, b3]),
        DenseAlignment([b16, b11, b8, b4]),
        DenseAlignment([b16, b11, b8, b5]),     # 54th
        DenseAlignment([b17, b9, b6, b3]),
        DenseAlignment([b17, b9, b6, b4]),
        DenseAlignment([b17, b9, b6, b5]),
        DenseAlignment([b17, b9, b7, b3]),
        DenseAlignment([b17, b9, b7, b4]),
        DenseAlignment([b17, b9, b7, b5]),
        DenseAlignment([b17, b9, b8, b3]),
        DenseAlignment([b17, b9, b8, b4]),
        DenseAlignment([b17, b9, b8, b5]),      # 63rd
        DenseAlignment([b17, b10, b6, b3]),
        DenseAlignment([b17, b10, b6, b4]),
        DenseAlignment([b17, b10, b6, b5]),
        DenseAlignment([b17, b10, b7, b3]),
        DenseAlignment([b17, b10, b7, b4]),
        DenseAlignment([b17, b10, b7, b5]),
        DenseAlignment([b17, b10, b8, b3]),
        DenseAlignment([b17, b10, b8, b4]),
        DenseAlignment([b17, b10, b8, b5]),     # 72nd
        DenseAlignment([b17, b11, b6, b3]),
        DenseAlignment([b17, b11, b6, b4]),
        DenseAlignment([b17, b11, b6, b5]),
        DenseAlignment([b17, b11, b7, b3]),
        DenseAlignment([b17, b11, b7, b4]),
        DenseAlignment([b17, b11, b7, b5]),
        DenseAlignment([b17, b11, b8, b3]),
        DenseAlignment([b17, b11, b8, b4]),
        DenseAlignment([b17, b11, b8, b5])]     # 81st
                                                # 3^4 = 81 = number of
                                                # quartets
class QuarterGeneratorTests(QuarterGeneratorSetUp):
    """Tests for quartet_generator function.
    """
    ## THIS SINGLE TEST BELOW CHECKS WHAT HAPPENS IF INPUT IS INVALID
    def test_invalid_input(self):
        """quartet_generator: should raise QuartetGeneratorError
        """
        g_no_DAs = quartet_generator()
        self.assertRaises(QuartetGeneratorError, g_no_DAs.next)
        
        g_two_DAs = quartet_generator(\
        DenseAlignment('11'), DenseAlignment('^^'))
        self.assertRaises(QuartetGeneratorError, g_two_DAs.next)
        
        g_three_DAs = quartet_generator(\
        DenseAlignment('11'), DenseAlignment('11'), DenseAlignment('77'))
        self.assertRaises(QuartetGeneratorError, g_three_DAs.next)
        
        g_one_ALN = quartet_generator(Alignment('T'))
        self.assertRaises(QuartetGeneratorError, g_one_ALN.next)
        
        g_four_ALNs = quartet_generator(Alignment('T'), Alignment('U'),\
        Alignment('G'), Alignment('C'))
        self.assertRaises(QuartetGeneratorError, g_four_ALNs.next)
        
        g_one_STR = quartet_generator('()*&')
        self.assertRaises(QuartetGeneratorError, g_one_STR.next)
        
        g_two_STRs = quartet_generator('ACGU', '()*&')
        self.assertRaises(QuartetGeneratorError, g_two_STRs.next)
        
        g_three_STRs = quartet_generator('ACGU', 'ACGU', 'ACGU')
        self.assertRaises(QuartetGeneratorError, g_three_STRs.next)
        
        g_four_STRs = quartet_generator('ACG?', 'ACG!', '&CGU', 'A6GU')
        self.assertRaises(QuartetGeneratorError, g_four_STRs.next)

        # If there are some duplicate sequences (duplicate names) in input, 
        # the QuartetGeneratorError is raised.
        g_FOUR_DAs_dupl = quartet_generator(\
        self.da1, self.da1, self.da1, self.da1)
        self.assertRaises(QuartetGeneratorError, g_FOUR_DAs_dupl.next)

        # If after gap removal there are no positions in sequences...
        # ...QuartetGeneratorError is raised.
        g_FOUR_DAs_no_seqs = quartet_generator(\
        self.da1, self.da2, self.da3, self.da4)
        self.assertRaises(QuartetGeneratorError, g_FOUR_DAs_no_seqs.next)
        
        g_ONE_DA_no_seqs = quartet_generator(DenseAlignment([\
        ByteSequence('----00-', Name='a'), ByteSequence('1111--1', Name='b')]))
        self.assertRaises(QuartetGeneratorError, g_ONE_DA_no_seqs.next)

    ## ALL TESTS BELOW CHECK WHAT HAPPENS IF INPUT IS VALID
    def test_correct_output(self):
        """quartet_generator: should create BinaryQuartet objects from correct input
        """
        g_DIGITS = quartet_generator(self.da_DIGITS)
        g_CHARS = quartet_generator(self.da_CHARS)

        for smth in g_DIGITS:
            self.assertTrue(isinstance(smth, BinaryQuartet))

        for smth in g_DIGITS:
            self.assertTrue(isinstance(smth, BinaryQuartet))
            
    def test_single_dense_aln(self):
        """quartet_generator: should create all Quartets from 1 DenseAlignment
        """
        # Obviously, only small DenseAlignment object should be tested here.
        g1 = quartet_generator(self.da_DIGITS)
        
        list_of_quartets = [quartet.Quartet for quartet in g1]
        # list_of_quartet contains DenseAlignments 
        # (each representing a quartet of sequences)
        
        self.assertEqual(len(list_of_quartets), 5)  # 5 = nr of unique quartets
                                                    # made from 5 seqs

    def test_many_dense_alns(self):
        """quartet_generator: should create all Quartets from 4 DenseAlignments
        """
        # Obviously, only small DenseAlignment object should be tested here.
        g_FOUR_DAs = quartet_generator(self.da5, self.da1, self.da3, self.da4)
        list_of_quartets = [quartet.Quartet for quartet in g_FOUR_DAs]
        
        self.assertEqual(len(list_of_quartets), 81)  # 81 = nr of quartets made
                                                     # from 4 DenseAlignment
                                                     # objects each containing
                                                # 3 ByteSequence objects
                                                     # what gives = 3^4 = 
                                                     # 81 quartets.

class CheckInputTests(QuarterGeneratorSetUp):
    """Tests for check_input function.
    """
    def test_valid_input(self):
        """check_input: should return unchanged input if the input is valid
        """
        self.assertSameObj(check_input([self.da_DIGITS])[0], self.da_DIGITS)
    
    def test_invalid_input(self):
        """check_input: should raise QuartetGeneratorError on invalid input
        """
        self.assertRaises(QuartetGeneratorError, check_input, [])
        
        self.assertRaises(\
        QuartetGeneratorError, check_input, [DenseAlignment('11'), \
        DenseAlignment('11')])
        
        self.assertRaises(\
        QuartetGeneratorError, check_input, [DenseAlignment('11'), \
        DenseAlignment('11'), DenseAlignment('11')])
        
        self.assertRaises(\
        QuartetGeneratorError, check_input, Alignment('T'))
        
        self.assertRaises(\
        QuartetGeneratorError, check_input, [Alignment('T')])
        
        self.assertRaises(\
        QuartetGeneratorError, check_input, \
        [Alignment('T'), Alignment('U'), Alignment('G'), Alignment('C')])
        
        self.assertRaises(\
        QuartetGeneratorError, check_input, ['AUGC'])
        
        self.assertRaises(\
        QuartetGeneratorError, check_input, ['7777', 'AUGC'])
        
        self.assertRaises(\
        QuartetGeneratorError, check_input, ['()*&', 'AUGC', 'AUGC'])

        self.assertRaises(\
        QuartetGeneratorError, check_input, \
        ['AUGC', 'AUGC', 'AUGC', 'AUGC'])

class SelectFromOneDATests(QuarterGeneratorSetUp):
    """Tests for select_from_one_DA function.
    """
    def test_select_from_one_DA_DIGITS(self):
        """select_from_one_DA: should return correct quartets made of DIGITS
        """
        items = [ByteSequence('11'), ByteSequence('00'), ByteSequence('10'),
        ByteSequence('01'), ByteSequence('11'), ByteSequence('01')]
        
        nr_of_quartets = 0
        for quartet in select_from_one_DA(items):
            self.assertTrue(isinstance(quartet, list))
            nr_of_quartets += 1
            
        self.assertEqual(nr_of_quartets, 15)    # 15 unique quartets can be
                                                # generated from 6 elements.
                                                
    def test_select_from_one_DA_CHARS(self):
        """select_from_one_DA: should return correct quartets made of CHARS
        """
        items = [ByteSequence('()'), ByteSequence('**'), ByteSequence('*^'),
        ByteSequence('(*'), ByteSequence('()'), ByteSequence('**')]
        
        nr_of_quartets = 0
        for quartet in select_from_one_DA(items):
            self.assertTrue(isinstance(quartet, list))
            nr_of_quartets += 1
            
        self.assertEqual(nr_of_quartets, 15)    # 15 unique quartets can be
                                                # generated from 6 elements.

class SelectFromFourDAsTests(QuarterGeneratorSetUp):
    """Tests for select_from_four_DAs function.
    """
    def test_select_from_four_DAs_DIGITS(self):
        """select_from_four_DAs: should return correct quartets made of DIGITS
        """
        items = \
        [self.da_DIGITS, self.da_DIGITS, self.da_DIGITS, self.da_DIGITS]
        nr_of_quartets = 0
        for quartet in select_from_four_DAs(items):
            self.assertTrue(isinstance(quartet, list))
            nr_of_quartets += 1
            
            for probably_a_list in quartet:
                self.assertTrue(isinstance(probably_a_list, list))
                for probably_a_str in probably_a_list:
                    self.assertTrue(isinstance(probably_a_str, str))
        
        self.assertEqual(nr_of_quartets, 625)   # 625 = 5^4
        
    def test_select_from_four_DAs_CHARS(self):
        """select_from_four_DAs: should return correct quartets made of CHARS
        """
        items = \
        [self.da_CHARS, self.da_CHARS, self.da_CHARS,self.da_CHARS]
        nr_of_quartets = 0
        for quartet in select_from_four_DAs(items):
            self.assertTrue(isinstance(quartet, list))
            nr_of_quartets += 1
            
            for probably_a_list in quartet:
                self.assertTrue(isinstance(probably_a_list, list))
                for probably_a_str in probably_a_list:
                    self.assertTrue(isinstance(probably_a_str, str))
        
        self.assertEqual(nr_of_quartets, 625)   # 625 = 5^4
        
class AverageQuartetTests(TestCaseBugFix):
    """Tests for AverageQuartet object.
    """
    def setUp(self):
        """Creates all datasets required by tests from AverageQuartetTests.
        """        
        self.q1 = Quartet(DenseAlignment([
        ByteSequence('11100'),
        ByteSequence('01011'), 
        ByteSequence('11011'),
        ByteSequence('00000')]))
        
        self.q2 = Quartet(DenseAlignment([
        ByteSequence('00000'),
        ByteSequence('01011'), 
        ByteSequence('11111'),
        ByteSequence('00000')]))
        
        self.q3 = Quartet(DenseAlignment([
        ByteSequence('10000'),
        ByteSequence('01111'), 
        ByteSequence('11001'),
        ByteSequence('00110')]))
        
        self.q4 = Quartet(DenseAlignment([
        ByteSequence('00100'),
        ByteSequence('01111'), 
        ByteSequence('11111'),
        ByteSequence('00100')]))
        
    def test__init__(self):
        """AverageQuartet.__init__: should initialize AverageQuartet correctly.
        """
        # Let's check private properties. as well.
        av_quartet = AverageQuartet()
        self.assertEqualItems(av_quartet._A_BCD, [0.0, 0])
        self.assertEqualItems(av_quartet._B_ACD, [0.0, 0])
        self.assertEqualItems(av_quartet._C_ABD, [0.0, 0])
        self.assertEqualItems(av_quartet._D_ABC, [0.0, 0])
        self.assertEqualItems(av_quartet._AB, [0.0, 0])
        self.assertEqualItems(av_quartet._AD, [0.0, 0])
        self.assertEqualItems(av_quartet._AC, [0.0, 0])
        
    def test_reset(self):
        """AverageQuartet.reset: should re-initialize AverageQuartet object.
        """
        av_quartet = AverageQuartet()
        self.q1.analyze()
        # self.q1 has just been analyzed, so the Params dict should be ready!
        av_quartet.update(self.q1)
        
        self.assertEqualItems(av_quartet._A_BCD, [1/5, 1])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 1])
        self.assertEqualItems(av_quartet._C_ABD, [0/5, 1])
        self.assertEqualItems(av_quartet._D_ABC, [1/5, 1])
        self.assertEqualItems(av_quartet._AB, [0/5, 1])
        self.assertEqualItems(av_quartet._AD, [2/5, 1])
        self.assertEqualItems(av_quartet._AC, [1/5, 1])
        
        self.q2.analyze()
        av_quartet.update(self.q2)
        # self.q1 has just been analyzed, so the Params dict should be ready!
        
        self.assertEqualItems(av_quartet._A_BCD, [1/5, 2])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 2])
        self.assertEqualItems(av_quartet._C_ABD, [2/5, 2])
        self.assertEqualItems(av_quartet._D_ABC, [1/5, 2])
        self.assertEqualItems(av_quartet._AB, [0/5, 2])
        self.assertEqualItems(av_quartet._AD, [5/5, 2])
        self.assertEqualItems(av_quartet._AC, [1/5, 2])
        
        av_quartet.calculateAverages()
        self.assertFloatEqual(av_quartet.Protrusion, 4/5/2/4)
        self.assertFloatEqual(av_quartet.SmallBox, 0/5/2)
        self.assertFloatEqual(av_quartet.MediumBox, 1/5/2)
        self.assertFloatEqual(av_quartet.LargeBox, 5/5/2)
        
        av_quartet.reset()
        # Now, all values should be set to 0.
        self.assertEqualItems(av_quartet._A_BCD, [0/5, 0])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 0])
        self.assertEqualItems(av_quartet._C_ABD, [0/5, 0])
        self.assertEqualItems(av_quartet._D_ABC, [0/5, 0])
        self.assertEqualItems(av_quartet._AB, [0, 0])
        self.assertEqualItems(av_quartet._AD, [0, 0])
        self.assertEqualItems(av_quartet._AC, [0, 0])
        
    def test_update_invalid(self):
        """AverageQuartet.update: should raise AverageQuartetError.
        """
        av_quartet = AverageQuartet()
        # self.q1 has not been analyzed (method analyze has not been used yet)
        # so the Params dict is empty!
        self.assertRaises(AverageQuartetError, av_quartet.update, self.q1)
        self.assertRaises(AverageQuartetError, av_quartet.update, '')
        self.assertRaises(\
        AverageQuartetError, av_quartet.update, DenseAlignment(['111']))
        
        self.assertRaises(TypeError, av_quartet.update)
        
    def test_update_valid(self):
        """AverageQuartet.update: should average Quartets' characteristics.
        """
        av_quartet = AverageQuartet()
        self.q1.analyze()
        # self.q1 has just been analyzed, so the Params dict should be ready!
        av_quartet.update(self.q1)
        
        self.assertEqualItems(av_quartet._A_BCD, [1/5, 1])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 1])
        self.assertEqualItems(av_quartet._C_ABD, [0/5, 1])
        self.assertEqualItems(av_quartet._D_ABC, [1/5, 1])
        self.assertEqualItems(av_quartet._AB, [0/5, 1])
        self.assertEqualItems(av_quartet._AD, [2/5, 1])
        self.assertEqualItems(av_quartet._AC, [1/5, 1])
        self.assertEqualItems(av_quartet._small_box, [0/5, 1])
        self.assertEqualItems(av_quartet._medium_box, [1/5, 1])
        self.assertEqualItems(av_quartet._large_box, [2/5, 1])
        self.assertEqualItems(av_quartet._d0, [0/5, 1])
        self.assertEqualItems(av_quartet._d1, [2/5, 1])
        self.assertEqualItems(av_quartet._d2, [3/5, 1])
        self.assertEqualItems(av_quartet._d3, [0, 0])
        self.assertEqualItems(av_quartet._d4, [0, 0])
        
        ##
        av_quartet.update(self.q1)
        
        self.assertEqualItems(av_quartet._A_BCD, [2/5, 2])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 2])
        self.assertEqualItems(av_quartet._C_ABD, [0/5, 2])
        self.assertEqualItems(av_quartet._D_ABC, [2/5, 2])
        self.assertEqualItems(av_quartet._AB, [0/5, 2])
        self.assertEqualItems(av_quartet._AD, [4/5, 2])
        self.assertEqualItems(av_quartet._AC, [2/5, 2])
        self.assertEqualItems(av_quartet._small_box, [0/5, 2])
        self.assertEqualItems(av_quartet._medium_box, [2/5, 2])
        self.assertEqualItems(av_quartet._large_box, [4/5, 2])
        self.assertEqualItems(av_quartet._d0, [0/5, 2])
        self.assertEqualItems(av_quartet._d1, [4/5, 2])
        self.assertEqualItems(av_quartet._d2, [6/5, 2])
        self.assertEqualItems(av_quartet._d3, [0, 0])
        self.assertEqualItems(av_quartet._d4, [0, 0])
        
        ##
        av_quartet.update(self.q1)
        
        self.assertEqualItems(av_quartet._A_BCD, [3/5, 3])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 3])
        self.assertEqualItems(av_quartet._C_ABD, [0/5, 3])
        self.assertEqualItems(av_quartet._D_ABC, [3/5, 3])
        self.assertEqualItems(av_quartet._AB, [0/5, 3])
        self.assertEqualItems(av_quartet._AD, [6/5, 3])
        self.assertEqualItems(av_quartet._AC, [3/5, 3])
        self.assertEqualItems(av_quartet._small_box, [0/5, 3])
        self.assertEqualItems(av_quartet._medium_box, [3/5, 3])
        self.assertEqualItems(av_quartet._large_box, [6/5, 3])
        self.assertEqualItems(av_quartet._d0, [0/5, 3])
        self.assertEqualItems(av_quartet._d1, [6/5, 3])
        self.assertEqualItems(av_quartet._d2, [9/5, 3])
        self.assertEqualItems(av_quartet._d3, [0, 0])
        self.assertEqualItems(av_quartet._d4, [0, 0])
                
        ##
        self.q2.analyze()
        # self.q2 has just been analyzed, so the Params dict should be ready!
        av_quartet.update(self.q2)
        
        self.assertEqualItems(av_quartet._A_BCD, [3/5, 4])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 4])
        self.assertEqualItems(av_quartet._C_ABD, [2/5, 4])
        self.assertEqualItems(av_quartet._D_ABC, [3/5, 4])
        self.assertEqualItems(av_quartet._AB, [0/5, 4])
        self.assertEqualItems(av_quartet._AD, [9/5, 4])
        self.assertEqualItems(av_quartet._AC, [3/5, 4])
        self.assertEqualItems(av_quartet._small_box, [0/5, 4])
        self.assertEqualItems(av_quartet._medium_box, [3/5, 4])
        self.assertEqualItems(av_quartet._large_box, [9/5, 4])
        self.assertEqualItems(av_quartet._d0, [0/5, 4])
        self.assertEqualItems(av_quartet._d1, [8/5, 4])
        self.assertEqualItems(av_quartet._d2, [12/5, 4])
        self.assertEqualItems(av_quartet._d3, [0, 0])
        self.assertEqualItems(av_quartet._d4, [0, 0])
        
        ##
        self.q3.analyze()
        # self.q3 has just been analyzed, so the Params dict should be ready!
        av_quartet.update(self.q3)
        
        self.assertEqualItems(av_quartet._A_BCD, [3/5, 5])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 5])
        self.assertEqualItems(av_quartet._C_ABD, [2/5, 5])
        self.assertEqualItems(av_quartet._D_ABC, [3/5, 5])
        self.assertEqualItems(av_quartet._AB, [0/5, 5])
        self.assertEqualItems(av_quartet._AD, [11/5, 5])
        self.assertEqualItems(av_quartet._AC, [6/5, 5])
        self.assertEqualItems(av_quartet._small_box, [0/5, 5])
        self.assertEqualItems(av_quartet._medium_box, [5/5, 5])
        self.assertEqualItems(av_quartet._large_box, [12/5, 5])
        self.assertEqualItems(av_quartet._d0, [0/5, 5])
        self.assertEqualItems(av_quartet._d1, [8/5, 5])
        self.assertEqualItems(av_quartet._d2, [17/5, 5])
        self.assertEqualItems(av_quartet._d3, [0, 0])
        self.assertEqualItems(av_quartet._d4, [0, 0])
        
        ##
        self.q4.analyze()
        # self.q4 has just been analyzed, so the Params dict should be ready!
        av_quartet.update(self.q4)
        
        self.assertEqualItems(av_quartet._A_BCD, [3/5, 6])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 6])
        self.assertEqualItems(av_quartet._C_ABD, [3/5, 6])
        self.assertEqualItems(av_quartet._D_ABC, [3/5, 6])
        self.assertEqualItems(av_quartet._AB, [0/5, 6])
        self.assertEqualItems(av_quartet._AD, [14/5, 6])
        self.assertEqualItems(av_quartet._AC, [6/5, 6])
        self.assertEqualItems(av_quartet._small_box, [0/5, 6])
        self.assertEqualItems(av_quartet._medium_box, [5/5, 6])
        self.assertEqualItems(av_quartet._large_box, [15/5, 6])
        self.assertEqualItems(av_quartet._d0, [1/5, 6])
        self.assertEqualItems(av_quartet._d1, [9/5, 6])
        self.assertEqualItems(av_quartet._d2, [20/5, 6])
        self.assertEqualItems(av_quartet._d3, [0, 0])
        self.assertEqualItems(av_quartet._d4, [0, 0])
        
        # If all above tests pass, so the one below should pass as well.
        av_quartet.calculateAverages()
        self.assertFloatEqual(av_quartet.Protrusion, 9/5/6/4)
        self.assertFloatEqual(av_quartet.SmallBox, 0/5/6)
        self.assertFloatEqual(av_quartet.MediumBox, 5/5/6)
        self.assertFloatEqual(av_quartet.LargeBox, 15/5/6)
        self.assertFloatEqual(av_quartet.d0, 1/5/6)
        self.assertFloatEqual(av_quartet.d1, 9/5/6)
        self.assertFloatEqual(av_quartet.d2, 20/5/6) # 6 times update() & 5 cols

class AverageBinaryQuartetTests(TestCaseBugFix):
    """Tests for AverageBinaryQuartet object.
    """
    def setUp(self):
        """Creates all datasets for tests from AverageBinaryQuartetTests.
        """
        self.q1 = Quartet(DenseAlignment([
        ByteSequence('11100'),
        ByteSequence('01011'), 
        ByteSequence('11011'),
        ByteSequence('00000')]))
        
        self.q2 = Quartet(DenseAlignment([
        ByteSequence('00000'),
        ByteSequence('01011'), 
        ByteSequence('11111'),
        ByteSequence('00000')]))
        
        self.q3 = Quartet(DenseAlignment([
        ByteSequence('10000'),
        ByteSequence('01111'), 
        ByteSequence('11001'),
        ByteSequence('00110')]))
        
        self.q4 = Quartet(DenseAlignment([
        ByteSequence('00100'),
        ByteSequence('01111'), 
        ByteSequence('11111'),
        ByteSequence('00100')]))
        
    def test__init__(self):
        """AverageBinaryQuartet.__init__: inits AverageBinaryQuartet
        """
        # Let's check private properties. as well.
        av_quartet = AverageBinaryQuartet()
        self.assertEqualItems(av_quartet._A_BCD, [0.0, 0])
        self.assertEqualItems(av_quartet._B_ACD, [0.0, 0])
        self.assertEqualItems(av_quartet._C_ABD, [0.0, 0])
        self.assertEqualItems(av_quartet._D_ABC, [0.0, 0])
        self.assertEqualItems(av_quartet._AB, [0.0, 0])
        self.assertEqualItems(av_quartet._AD, [0.0, 0])
        self.assertEqualItems(av_quartet._AC, [0.0, 0])
        
    def test_reset(self):
        """AverageBinaryQuartet.reset: re-inits AverageBinaryQuartet
        """
        av_quartet = AverageBinaryQuartet()
        self.q1.analyze()
        # self.q1 has just been analyzed, so the Params dict should be ready!
        av_quartet.update(self.q1)
        
        self.assertEqualItems(av_quartet._A_BCD, [1/5, 1])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 1])
        self.assertEqualItems(av_quartet._C_ABD, [0/5, 1])
        self.assertEqualItems(av_quartet._D_ABC, [1/5, 1])
        self.assertEqualItems(av_quartet._AB, [0/5, 1])
        self.assertEqualItems(av_quartet._AD, [2/5, 1])
        self.assertEqualItems(av_quartet._AC, [1/5, 1])
        
        self.q2.analyze()
        av_quartet.update(self.q2)
        # self.q1 has just been analyzed, so the Params dict should be ready!
        
        self.assertEqualItems(av_quartet._A_BCD, [1/5, 2])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 2])
        self.assertEqualItems(av_quartet._C_ABD, [2/5, 2])
        self.assertEqualItems(av_quartet._D_ABC, [1/5, 2])
        self.assertEqualItems(av_quartet._AB, [0/5, 2])
        self.assertEqualItems(av_quartet._AD, [5/5, 2])
        self.assertEqualItems(av_quartet._AC, [1/5, 2])
        
        av_quartet.calculateAverages()
        self.assertFloatEqual(av_quartet.Protrusion, 4/5/2/4)
        self.assertFloatEqual(av_quartet.SmallBox, 0/5/2)
        self.assertFloatEqual(av_quartet.MediumBox, 1/5/2)
        self.assertFloatEqual(av_quartet.LargeBox, 5/5/2)
        
        av_quartet.reset()
        # Now, all values should be set to 0.
        self.assertEqualItems(av_quartet._A_BCD, [0/5, 0])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 0])
        self.assertEqualItems(av_quartet._C_ABD, [0/5, 0])
        self.assertEqualItems(av_quartet._D_ABC, [0/5, 0])
        self.assertEqualItems(av_quartet._AB, [0, 0])
        self.assertEqualItems(av_quartet._AD, [0, 0])
        self.assertEqualItems(av_quartet._AC, [0, 0])
        
    def test_update_invalid(self):
        """AverageBinaryQuartet.update: should raise AverageQuartetError.
        """
        av_quartet = AverageBinaryQuartet()
        # self.q1 has not been analyzed (method analyze has not been used yet)
        # so the Params dict is empty!
        self.assertRaises(AverageQuartetError, av_quartet.update, self.q1)
        self.assertRaises(AverageQuartetError, av_quartet.update, '')
        self.assertRaises(\
        AverageQuartetError, av_quartet.update, DenseAlignment(['111']))
        
        self.assertRaises(TypeError, av_quartet.update)
        
    def test_update_valid(self):
        """AverageBinaryQuartet.update: should avg Quartets' charact.
        """
        av_quartet = AverageBinaryQuartet()
        self.q1.analyze()
        # self.q1 has just been analyzed, so the Params dict should be ready!
        av_quartet.update(self.q1)
        
        self.assertEqualItems(av_quartet._A_BCD, [1/5, 1])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 1])
        self.assertEqualItems(av_quartet._C_ABD, [0/5, 1])
        self.assertEqualItems(av_quartet._D_ABC, [1/5, 1])
        self.assertEqualItems(av_quartet._AB, [0/5, 1])
        self.assertEqualItems(av_quartet._AD, [2/5, 1])
        self.assertEqualItems(av_quartet._AC, [1/5, 1])
        self.assertEqualItems(av_quartet._small_box, [0/5, 1])
        self.assertEqualItems(av_quartet._medium_box, [1/5, 1])
        self.assertEqualItems(av_quartet._large_box, [2/5, 1])
        self.assertEqualItems(av_quartet._d0, [0/5, 1])
        self.assertEqualItems(av_quartet._d1, [2/5, 1])
        self.assertEqualItems(av_quartet._d2, [3/5, 1])
        self.assertEqualItems(av_quartet._d3, [0, 0])
        self.assertEqualItems(av_quartet._d4, [0, 0])
        
        ##
        av_quartet.update(self.q1)
        
        self.assertEqualItems(av_quartet._A_BCD, [2/5, 2])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 2])
        self.assertEqualItems(av_quartet._C_ABD, [0/5, 2])
        self.assertEqualItems(av_quartet._D_ABC, [2/5, 2])
        self.assertEqualItems(av_quartet._AB, [0/5, 2])
        self.assertEqualItems(av_quartet._AD, [4/5, 2])
        self.assertEqualItems(av_quartet._AC, [2/5, 2])
        self.assertEqualItems(av_quartet._small_box, [0/5, 2])
        self.assertEqualItems(av_quartet._medium_box, [2/5, 2])
        self.assertEqualItems(av_quartet._large_box, [4/5, 2])
        self.assertEqualItems(av_quartet._d0, [0/5, 2])
        self.assertEqualItems(av_quartet._d1, [4/5, 2])
        self.assertEqualItems(av_quartet._d2, [6/5, 2])
        self.assertEqualItems(av_quartet._d3, [0, 0])
        self.assertEqualItems(av_quartet._d4, [0, 0])
    
        ##
        av_quartet.update(self.q1)
        
        self.assertEqualItems(av_quartet._A_BCD, [3/5, 3])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 3])
        self.assertEqualItems(av_quartet._C_ABD, [0/5, 3])
        self.assertEqualItems(av_quartet._D_ABC, [3/5, 3])
        self.assertEqualItems(av_quartet._AB, [0/5, 3])
        self.assertEqualItems(av_quartet._AD, [6/5, 3])
        self.assertEqualItems(av_quartet._AC, [3/5, 3])
        self.assertEqualItems(av_quartet._small_box, [0/5, 3])
        self.assertEqualItems(av_quartet._medium_box, [3/5, 3])
        self.assertEqualItems(av_quartet._large_box, [6/5, 3])
        self.assertEqualItems(av_quartet._d0, [0/5, 3])
        self.assertEqualItems(av_quartet._d1, [6/5, 3])
        self.assertEqualItems(av_quartet._d2, [9/5, 3])
        self.assertEqualItems(av_quartet._d3, [0, 0])
        self.assertEqualItems(av_quartet._d4, [0, 0])
                
        ##
        self.q2.analyze()
        # self.q2 has just been analyzed, so the Params dict should be ready!
        av_quartet.update(self.q2)
        
        self.assertEqualItems(av_quartet._A_BCD, [3/5, 4])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 4])
        self.assertEqualItems(av_quartet._C_ABD, [2/5, 4])
        self.assertEqualItems(av_quartet._D_ABC, [3/5, 4])
        self.assertEqualItems(av_quartet._AB, [0/5, 4])
        self.assertEqualItems(av_quartet._AD, [9/5, 4])
        self.assertEqualItems(av_quartet._AC, [3/5, 4])
        self.assertEqualItems(av_quartet._small_box, [0/5, 4])
        self.assertEqualItems(av_quartet._medium_box, [3/5, 4])
        self.assertEqualItems(av_quartet._large_box, [9/5, 4])
        self.assertEqualItems(av_quartet._d0, [0/5, 4])
        self.assertEqualItems(av_quartet._d1, [8/5, 4])
        self.assertEqualItems(av_quartet._d2, [12/5, 4])
        self.assertEqualItems(av_quartet._d3, [0, 0])
        self.assertEqualItems(av_quartet._d4, [0, 0])
        
        ##
        self.q3.analyze()
        # self.q3 has just been analyzed, so the Params dict should be ready!
        av_quartet.update(self.q3)
        
        self.assertEqualItems(av_quartet._A_BCD, [3/5, 5])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 5])
        self.assertEqualItems(av_quartet._C_ABD, [2/5, 5])
        self.assertEqualItems(av_quartet._D_ABC, [3/5, 5])
        self.assertEqualItems(av_quartet._AB, [0/5, 5])
        self.assertEqualItems(av_quartet._AD, [11/5, 5])
        self.assertEqualItems(av_quartet._AC, [6/5, 5])
        self.assertEqualItems(av_quartet._small_box, [0/5, 5])
        self.assertEqualItems(av_quartet._medium_box, [5/5, 5])
        self.assertEqualItems(av_quartet._large_box, [12/5, 5])
        self.assertEqualItems(av_quartet._d0, [0/5, 5])
        self.assertEqualItems(av_quartet._d1, [8/5, 5])
        self.assertEqualItems(av_quartet._d2, [17/5, 5])
        self.assertEqualItems(av_quartet._d3, [0, 0])
        self.assertEqualItems(av_quartet._d4, [0, 0])
        
        ##
        self.q4.analyze()
        # self.q4 has just been analyzed, so the Params dict should be ready!
        av_quartet.update(self.q4)
        
        self.assertEqualItems(av_quartet._A_BCD, [3/5, 6])
        self.assertEqualItems(av_quartet._B_ACD, [0/5, 6])
        self.assertEqualItems(av_quartet._C_ABD, [3/5, 6])
        self.assertEqualItems(av_quartet._D_ABC, [3/5, 6])
        self.assertEqualItems(av_quartet._AB, [0/5, 6])
        self.assertEqualItems(av_quartet._AD, [14/5, 6])
        self.assertEqualItems(av_quartet._AC, [6/5, 6])
        self.assertEqualItems(av_quartet._small_box, [0/5, 6])
        self.assertEqualItems(av_quartet._medium_box, [5/5, 6])
        self.assertEqualItems(av_quartet._large_box, [15/5, 6])
        self.assertEqualItems(av_quartet._d0, [1/5, 6])
        self.assertEqualItems(av_quartet._d1, [9/5, 6])
        self.assertEqualItems(av_quartet._d2, [20/5, 6])
        self.assertEqualItems(av_quartet._d3, [0, 0])
        self.assertEqualItems(av_quartet._d4, [0, 0])
        
        # If all above tests pass, so the one below should pass as well.
        av_quartet.calculateAverages()
        self.assertFloatEqual(av_quartet.Protrusion, 9/5/6/4)
        self.assertFloatEqual(av_quartet.SmallBox, 0/5/6)
        self.assertFloatEqual(av_quartet.MediumBox, 5/5/6)
        self.assertFloatEqual(av_quartet.LargeBox, 15/5/6)
        self.assertFloatEqual(av_quartet.d0, 1/5/6)
        self.assertFloatEqual(av_quartet.d1, 9/5/6)
        self.assertFloatEqual(av_quartet.d2, 20/5/6) # 6 times update() & 5 cols

class AverageQuaternaryQuartetTests(TestCaseBugFix):
    """Tests for AverageQuaternaryQuartet object.
    """
    def setUp(self):
        """Creates all datasets for tests from AverageQuaternaryQuartetTests.
        """
        b1 = ByteSequence('112233441123241214', Name='b1')
        b2 = ByteSequence('112233111141124121', Name='b2')
        b3 = ByteSequence('444444112223112333', Name='b3')
        b4 = ByteSequence('444433111312424342', Name='b4')
        b5 = ByteSequence('212233141111111111', Name='b5')
        b6 = ByteSequence('232233441444422231', Name='b6')

        
        self.q1 = QuaternaryQuartet(DenseAlignment([b1, b2, b3, b4]))
        self.q2 = QuaternaryQuartet(DenseAlignment([b2, b3, b4, b5]))
        self.q3 = QuaternaryQuartet(DenseAlignment([b3, b4, b5, b6]))
        
    def test__init__(self):
        """AverageQuaternaryQuartet.__init__: inits AverageQuaternaryQuartet.
        """
        # Let's check private properties. as well.
        av_quartet = AverageQuaternaryQuartet()
        self.assertEqualItems(av_quartet._A_BCD, [0.0, 0])
        self.assertEqualItems(av_quartet._B_ACD, [0.0, 0])
        self.assertEqualItems(av_quartet._C_ABD, [0.0, 0])
        self.assertEqualItems(av_quartet._D_ABC, [0.0, 0])
        self.assertEqualItems(av_quartet._AB, [0.0, 0])
        self.assertEqualItems(av_quartet._AD, [0.0, 0])
        self.assertEqualItems(av_quartet._AC, [0.0, 0])
        
    def test_reset(self):
        """AverageQuaternaryQuartet.reset: re-inits AverageQuaternaryQuartet.
        """
        av_quartet = AverageQuaternaryQuartet()
        self.q1.analyze()
        # self.q1 has just been analyzed, so the Params dict should be ready!
        av_quartet.update(self.q1)
        
        self.assertEqualItems(av_quartet._A_BCD, [2/18, 1])
        self.assertEqualItems(av_quartet._B_ACD, [0/18, 1])
        self.assertEqualItems(av_quartet._C_ABD, [3/18, 1])
        self.assertEqualItems(av_quartet._D_ABC, [0/18, 1])
        self.assertEqualItems(av_quartet._AB, [4/18, 1])
        self.assertEqualItems(av_quartet._AD, [0/18, 1])
        self.assertEqualItems(av_quartet._AC, [0/18, 1])
        
        self.q2.analyze()
        av_quartet.update(self.q2)
        # self.q1 has just been analyzed, so the Params dict should be ready!
        
        self.assertEqualItems(av_quartet._A_BCD, [2/18, 2])
        self.assertEqualItems(av_quartet._B_ACD, [3/18, 2])
        self.assertEqualItems(av_quartet._C_ABD, [4/18, 2])
        self.assertEqualItems(av_quartet._D_ABC, [1/18, 2])
        self.assertEqualItems(av_quartet._AB, [4/18, 2])
        self.assertEqualItems(av_quartet._AD, [4/18, 2])
        self.assertEqualItems(av_quartet._AC, [1/18, 2])
        
        av_quartet.calculateAverages()
        self.assertFloatEqual(av_quartet.Protrusion, 10/18/2/4)
        self.assertFloatEqual(av_quartet.SmallBox, 0/18/2)
        self.assertFloatEqual(av_quartet.MediumBox, 1/18/2)
        self.assertFloatEqual(av_quartet.LargeBox, 8/18/2)
        
        av_quartet.reset()
        # Now, all values should be set to 0.
        self.assertEqualItems(av_quartet._A_BCD, [0/18, 0])
        self.assertEqualItems(av_quartet._B_ACD, [0/18, 0])
        self.assertEqualItems(av_quartet._C_ABD, [0/18, 0])
        self.assertEqualItems(av_quartet._D_ABC, [0/18, 0])
        self.assertEqualItems(av_quartet._AB, [0, 0])
        self.assertEqualItems(av_quartet._AD, [0, 0])
        self.assertEqualItems(av_quartet._AC, [0, 0])
        
    def test_update_invalid(self):
        """AverageQuaternaryQuartet.update:raises AverageQuartetError
        """
        av_quartet = AverageQuaternaryQuartet()
        # self.q1 has not been analyzed (method analyze has not been used yet)
        # so the Params dict is empty!
        self.assertRaises(AverageQuartetError, av_quartet.update, self.q1)
        self.assertRaises(AverageQuartetError, av_quartet.update, '')
        self.assertRaises(\
        AverageQuartetError, av_quartet.update, DenseAlignment(['111']))
        
        self.assertRaises(TypeError, av_quartet.update)
        
    def test_update_valid(self):
        """AverageQuaternaryQuartet.update:shall avg Quartets charact
        """
        av_quartet = AverageQuaternaryQuartet()
        self.q1.analyze()
        # self.q1 has just been analyzed, so the Params dict should be ready!
        av_quartet.update(self.q1)
        
        self.assertEqualItems(av_quartet._A_BCD, [2/18, 1])
        self.assertEqualItems(av_quartet._B_ACD, [0/18, 1])
        self.assertEqualItems(av_quartet._C_ABD, [3/18, 1])
        self.assertEqualItems(av_quartet._D_ABC, [0/18, 1])
        self.assertEqualItems(av_quartet._AB, [4/18, 1])
        self.assertEqualItems(av_quartet._AD, [0/18, 1])
        self.assertEqualItems(av_quartet._AC, [0/18, 1])
        self.assertEqualItems(av_quartet._small_box, [0/18, 1])
        self.assertEqualItems(av_quartet._medium_box, [0/18, 1])
        self.assertEqualItems(av_quartet._large_box, [4/18, 1])
        self.assertEqualItems(av_quartet._d0, [0/18, 1])
        self.assertEqualItems(av_quartet._d1, [5/18, 1])
        self.assertEqualItems(av_quartet._d2, [4/18, 1])
        
        self.assertEqualItems(av_quartet._AB_C_D, [1/18, 1])
        self.assertEqualItems(av_quartet._AC_B_D, [2/18, 1])
        self.assertEqualItems(av_quartet._AD_B_C, [0/18, 1])
        self.assertEqualItems(av_quartet._BC_A_D, [1/18, 1])
        self.assertEqualItems(av_quartet._BD_A_C, [2/18, 1])
        self.assertEqualItems(av_quartet._CD_A_B, [1/18, 1])
        self.assertEqualItems(av_quartet._small_tetra, [0/18, 1])
        self.assertEqualItems(av_quartet._medium_tetra, [1/18, 1])
        self.assertEqualItems(av_quartet._large_tetra, [2/18, 1])
        self.assertEqualItems(av_quartet._d3, [7/18, 1])
        self.assertEqualItems(av_quartet._d4, [2/18, 1])
        
        ##
        av_quartet.update(self.q1)
        
        self.assertEqualItems(av_quartet._A_BCD, [4/18, 2])
        self.assertEqualItems(av_quartet._B_ACD, [0/18, 2])
        self.assertEqualItems(av_quartet._C_ABD, [6/18, 2])
        self.assertEqualItems(av_quartet._D_ABC, [0/18, 2])
        self.assertEqualItems(av_quartet._AB, [8/18, 2])
        self.assertEqualItems(av_quartet._AD, [0/18, 2])
        self.assertEqualItems(av_quartet._AC, [0/18, 2])
        self.assertEqualItems(av_quartet._small_box, [0/18, 2])
        self.assertEqualItems(av_quartet._medium_box, [0/18, 2])
        self.assertEqualItems(av_quartet._large_box, [8/18, 2])
        self.assertEqualItems(av_quartet._d0, [0/18, 2])
        self.assertEqualItems(av_quartet._d1, [10/18, 2])
        self.assertEqualItems(av_quartet._d2, [8/18, 2])
        
        self.assertEqualItems(av_quartet._AB_C_D, [2/18, 2])
        self.assertEqualItems(av_quartet._AC_B_D, [4/18, 2])
        self.assertEqualItems(av_quartet._AD_B_C, [0/18, 2])
        self.assertEqualItems(av_quartet._BC_A_D, [2/18, 2])
        self.assertEqualItems(av_quartet._BD_A_C, [4/18, 2])
        self.assertEqualItems(av_quartet._CD_A_B, [2/18, 2])
        self.assertEqualItems(av_quartet._small_tetra, [0/18, 2])
        self.assertEqualItems(av_quartet._medium_tetra, [2/18, 2])
        self.assertEqualItems(av_quartet._large_tetra, [4/18, 2])
        self.assertEqualItems(av_quartet._d3, [14/18, 2])
        self.assertEqualItems(av_quartet._d4, [4/18, 2])
        ##
        av_quartet.update(self.q1)
        
        self.assertEqualItems(av_quartet._A_BCD, [6/18, 3])
        self.assertEqualItems(av_quartet._B_ACD, [0/18, 3])
        self.assertEqualItems(av_quartet._C_ABD, [9/18, 3])
        self.assertEqualItems(av_quartet._D_ABC, [0/18, 3])
        self.assertEqualItems(av_quartet._AB, [12/18, 3])
        self.assertEqualItems(av_quartet._AD, [0/18, 3])
        self.assertEqualItems(av_quartet._AC, [0/18, 3])
        self.assertEqualItems(av_quartet._small_box, [0/18, 3])
        self.assertEqualItems(av_quartet._medium_box, [0/18, 3])
        self.assertEqualItems(av_quartet._large_box, [12/18, 3])
        self.assertEqualItems(av_quartet._d0, [0/18, 3])
        self.assertEqualItems(av_quartet._d1, [15/18, 3])
        self.assertEqualItems(av_quartet._d2, [12/18, 3])
        
        self.assertEqualItems(av_quartet._AB_C_D, [3/18, 3])
        self.assertEqualItems(av_quartet._AC_B_D, [6/18, 3])
        self.assertEqualItems(av_quartet._AD_B_C, [0/18, 3])
        self.assertEqualItems(av_quartet._BC_A_D, [3/18, 3])
        self.assertEqualItems(av_quartet._BD_A_C, [6/18, 3])
        self.assertEqualItems(av_quartet._CD_A_B, [3/18, 3])
        self.assertEqualItems(av_quartet._small_tetra, [0/18, 3])
        self.assertEqualItems(av_quartet._medium_tetra, [3/18, 3])
        self.assertEqualItems(av_quartet._large_tetra, [6/18, 3])
        self.assertEqualItems(av_quartet._d3, [21/18, 3])
        self.assertEqualItems(av_quartet._d4, [6/18, 3])
        ##
        self.q2.analyze()
        av_quartet.update(self.q2)
        
        self.assertEqualItems(av_quartet._A_BCD, [6/18, 4])
        self.assertEqualItems(av_quartet._B_ACD, [3/18, 4])
        self.assertEqualItems(av_quartet._C_ABD, [10/18, 4])
        self.assertEqualItems(av_quartet._D_ABC, [1/18, 4])
        self.assertEqualItems(av_quartet._AB, [12/18, 4])
        self.assertEqualItems(av_quartet._AD, [4/18, 4])
        self.assertEqualItems(av_quartet._AC, [1/18, 4])
        self.assertEqualItems(av_quartet._small_box, [0/18, 4])
        self.assertEqualItems(av_quartet._medium_box, [1/18, 4])
        self.assertEqualItems(av_quartet._large_box, [16/18, 4])
        self.assertEqualItems(av_quartet._d0, [1/18, 4])
        self.assertEqualItems(av_quartet._d1, [20/18, 4])
        self.assertEqualItems(av_quartet._d2, [17/18, 4])
        
        self.assertEqualItems(av_quartet._AB_C_D, [3/18, 4])
        self.assertEqualItems(av_quartet._AC_B_D, [7/18, 4])
        self.assertEqualItems(av_quartet._AD_B_C, [3/18, 4])
        self.assertEqualItems(av_quartet._BC_A_D, [4/18, 4])
        self.assertEqualItems(av_quartet._BD_A_C, [6/18, 4])
        self.assertEqualItems(av_quartet._CD_A_B, [4/18, 4])
        self.assertEqualItems(av_quartet._small_tetra, [0/18, 4])
        self.assertEqualItems(av_quartet._medium_tetra, [4/18, 4])
        self.assertEqualItems(av_quartet._large_tetra, [9/18, 4])
        self.assertEqualItems(av_quartet._d3, [27/18, 4])
        self.assertEqualItems(av_quartet._d4, [7/18, 4])
        
        ##
        self.q3.analyze()
        # self.q3 has just been analyzed, so the Params dict should be ready!
        av_quartet.update(self.q3)
        
        self.assertEqualItems(av_quartet._A_BCD, [9/18, 5])
        self.assertEqualItems(av_quartet._B_ACD, [3/18, 5])
        self.assertEqualItems(av_quartet._C_ABD, [10/18, 5])
        self.assertEqualItems(av_quartet._D_ABC, [2/18, 5])
        self.assertEqualItems(av_quartet._AB, [16/18, 5])
        self.assertEqualItems(av_quartet._AD, [4/18, 5])
        self.assertEqualItems(av_quartet._AC, [3/18, 5])
        self.assertEqualItems(av_quartet._small_box, [0/18, 5])
        self.assertEqualItems(av_quartet._medium_box, [3/18, 5])
        self.assertEqualItems(av_quartet._large_box, [20/18, 5])
        self.assertEqualItems(av_quartet._d0, [1/18, 5])
        self.assertEqualItems(av_quartet._d1, [24/18, 5])
        self.assertEqualItems(av_quartet._d2, [23/18, 5])
        
        self.assertEqualItems(av_quartet._AB_C_D, [5/18, 5])
        self.assertEqualItems(av_quartet._AC_B_D, [7/18, 5])
        self.assertEqualItems(av_quartet._AD_B_C, [5/18, 5])
        self.assertEqualItems(av_quartet._BC_A_D, [5/18, 5])
        self.assertEqualItems(av_quartet._BD_A_C, [6/18, 5])
        self.assertEqualItems(av_quartet._CD_A_B, [5/18, 5])
        self.assertEqualItems(av_quartet._small_tetra, [0/18, 5])
        self.assertEqualItems(av_quartet._medium_tetra, [5/18, 5])
        self.assertEqualItems(av_quartet._large_tetra, [11/18, 5])
        self.assertEqualItems(av_quartet._d3, [33/18, 5])
        self.assertEqualItems(av_quartet._d4, [9/18, 5])
        
        # If all above tests pass, so the one below should pass as well.
        av_quartet.calculateAverages()
        self.assertFloatEqual(av_quartet.SmallBox, 0/18/5)  # 5 times update()
        self.assertFloatEqual(av_quartet.MediumBox, 3/18/5) # 18 columns
        self.assertFloatEqual(av_quartet.LargeBox, 20/18/5)
        self.assertFloatEqual(av_quartet.SmallTetra, 0/18/5)
        self.assertFloatEqual(av_quartet.MediumTetra, 5/18/5)
        self.assertFloatEqual(av_quartet.LargeTetra, 11/18/5)
        # They are None because for more than 1 quartet individual protrusions
        # cannot be distinguished anymore...
        self.assertTrue(av_quartet.Alpha is None)
        self.assertTrue(av_quartet.Beta is None)
        self.assertTrue(av_quartet.Gamma is None)
        self.assertTrue(av_quartet.Delta is None)
        # ...therefore we only get the length of an average protrusion.
        self.assertFloatEqual(av_quartet.Protrusion, 24/18/5/4)
        self.assertEqualItems(av_quartet._d0, [1/18, 5])
        self.assertEqualItems(av_quartet._d1, [24/18, 5])
        self.assertEqualItems(av_quartet._d2, [23/18, 5])
        self.assertEqualItems(av_quartet._d3, [33/18, 5])
        self.assertEqualItems(av_quartet._d4, [9/18, 5])
        
        self.assertFloatEqual(av_quartet.d0, 1/18/5)
        self.assertFloatEqual(av_quartet.d1, 24/18/5)
        self.assertFloatEqual(av_quartet.d2, 23/18/5)
        self.assertFloatEqual(av_quartet.d3, 33/18/5)
        self.assertFloatEqual(av_quartet.d4, 9/18/5)

if __name__ == '__main__':
    main()
